from __future__ import annotations
import tempfile, logging, os, rtoml, re
import zipfile, shutil
from copy import deepcopy
from pathlib import Path
from itertools import combinations, product
from typing import Callable, Tuple, Union, List, Dict
from unittest import mock

import networkx as nx
from tqdm import tqdm

import numpy as np
from numpy.typing import ArrayLike
from scipy.spatial.distance import cdist
import MDAnalysis as mda
import MDAnalysis.transformations

import chilife
from .protein_utils import dihedral_defs, local_mx, sort_pdb, mutate, get_min_topol
from .ProteinIC import ProteinIC
from .scoring import get_lj_rep, GAS_CONST, get_lj_params, reweight_rotamers
from .numba_utils import get_delta_r, normdist
from .SpinLabel import SpinLabel
from .RotamerEnsemble import RotamerEnsemble
from .SpinLabelTraj import SpinLabelTraj

logging.captureWarnings(True)

# Define useful global variables
SUPPORTED_BB_LABELS = ("R1C",)
DATA_DIR = Path(__file__).parent.absolute() / "data/"
RL_DIR = Path(__file__).parent.absolute() / "data/rotamer_libraries/"

with open(RL_DIR / 'additional_rotlib_dirs.txt', 'r') as f:
    USER_RL_DIR = [Path(x.strip()) for x in f.readlines()]

with open(RL_DIR / "defaults.toml", "r") as f:
    rotlib_defaults = rtoml.load(f)

USER_LIBRARIES = {f.name[:-11] for f in (RL_DIR / "user_rotlibs").glob("*rotlib.npz")}
USER_dLIBRARIES = {f.name[:-12] for f in (RL_DIR / "user_rotlibs").glob("*drotlib.zip")}
USER_dLIBRARIES = USER_dLIBRARIES | {f.name[:-15] for f in (RL_DIR / "user_rotlibs").glob("*drotlib.zip")}
SUPPORTED_RESIDUES = set(dihedral_defs.keys())
[SUPPORTED_RESIDUES.remove(lab) for lab in ("CYR1", "MTN")]


def distance_distribution(
        *args: SpinLabel,
        r: ArrayLike = None,
        sigma: float = 1.0,
        use_spin_centers: bool = True,
        dependent: bool = False,
        uq: bool = False,
) -> np.ndarray:
    """Calculates total distribution of spin-spin distances among an arbitrary number of spin labels, using the
    distance range ``r`` (in angstrom). The distance distribution is obtained by summing over all pair distance
    distributions. These in turn are calculated by summing over rotamer pairs with the appropriate weights. For each
    rotamer pair, the distance distribution is either just the distance between the ``spin_center`` coordinates of two
    labels (if ``spin_populations=False``) or the weighted sum over all pairs of spn-bearing atoms
    (``spin_populations=True``). The resulting distance histogram is convolved with a normal distribution with a
    standard deviation ``sigma`` .

    Parameters
    ----------
    *args : SpinLabel
        Any number of spin label objects.
    r : ArrayLike
        Evenly spaced array of distances (in angstrom) to calculate the distance distribution over.
    sigma : float
        The standard deviation of the normal distribution used in convolution with the distance histogram, in angstrom.
        Default is 1.
    use_spin_centers : bool
        If False, distances are computed between spin centers. If True, distances are computed by summing over
        the distributed spin density on spin-bearing atoms on the labels.
    dependent: bool
        Consider the (clash) effects of spin label rotamers on each other.
    uq : bool
        Perform uncertainty analysis (experimental)

    Returns
    -------
    P : np.ndarray
        Predicted distance distribution, in 1/angstrom
    """

    # Allow r to be passed as last non-keyword argument
    if r is None and np.ndim(args[-1]) != 0:
        r = args[-1]
        args = args[:-1]

    if len(args) < 2:
        raise TypeError('At least two spin label objects are required.')

    if r is None:
        raise TypeError('Keyword argument r with distance domain vector is missing.')

    if any(not hasattr(arg, atr) for arg in args for atr in ["spin_coords", "spin_centers", "weights"]):
        raise TypeError(
            "Arguments other than spin labels must be passed as a keyword arguments."
        )

    r = np.asarray(r)

    if any(isinstance(SL, SpinLabelTraj) for SL in args):

        P = traj_dd(*args, r=r, sigma=sigma)
        return P

    elif uq:

        Ps = []
        n_boots = uq if uq > 1 else 100
        for i in range(n_boots):
            dummy_labels = []
            for SL in args:
                idxs = np.random.choice(len(SL), len(SL))

                dummy_SL = mock.Mock()
                dummy_SL.spin_coords = np.atleast_2d(SL.spin_coords[idxs])
                dummy_SL.spin_centers = np.atleast_2d(SL.spin_centers[idxs])
                dummy_SL.spin_weights = SL.spin_weights
                dummy_SL.weights = SL.weights[idxs]
                dummy_SL.weights /= dummy_SL.weights.sum()
                dummy_labels.append(dummy_SL)

            Ps.append(pair_dd(*dummy_labels, r=r, sigma=sigma, use_spin_centers=use_spin_centers))
        Ps = np.array(Ps)
        return Ps

    else:

        P = pair_dd(*args, r=r, sigma=sigma, use_spin_centers=use_spin_centers, dependent=dependent)
        return P


def pair_dd(*args, r: ArrayLike, sigma: float = 1.0, use_spin_centers: bool = True, dependent=False) -> np.ndarray:
    """Obtain the total pairwise spin-spin distance distribution over ``r`` for a list of spin labels.
    The distribution is calculated by convolving the weighted histogram of pairwise spin-spin
    distances with a normal distribution with standard deviation ``sigma``.

    Parameters
    ----------
    *args : SpinLabel
        SpinLabels to use when calculating the distance distribution
    r : ArrayLike
        Distance domain vector, in angstrom
    sigma : float
         Standard deviation of normal distribution used for convolution, in angstrom
    use_spin_centers : bool
        If False, distances are computed between spin centers. If True, distances are computed by summing over
        the distributed spin density on spin-bearing atoms on the labels.
    dependent: bool
        Consider the (clash) effects of spin label rotamers on each other.

    Returns
    -------
    P : np.ndarray
        Predicted normalized distance distribution, in units of 1/angstrom

    """
    # Calculate pairwise distances and weights
    SL_Pairs = combinations(args, 2)
    weights, distances = [], []
    for SL1, SL2 in SL_Pairs:
        if use_spin_centers:
            coords1 = SL1.spin_centers
            coords2 = SL2.spin_centers
            weights1 = SL1.weights
            weights2 = SL2.weights
        else:
            coords1 = SL1.spin_coords.reshape(-1, 3)
            coords2 = SL2.spin_coords.reshape(-1, 3)
            weights1 = np.outer(SL1.weights, SL1.spin_weights).flatten()
            weights2 = np.outer(SL2.weights, SL2.spin_weights).flatten()

        distances.append(cdist(coords1, coords2).flatten())
        weights.append(np.outer(weights1, weights2).flatten())

        if dependent:
            nrot1, nrot2 = len(SL1), len(SL2)
            nat1, nat2 = len(SL1.side_chain_idx), len(SL2.side_chain_idx)

            rot_coords1 = SL1.coords[:, SL1.side_chain_idx].reshape(-1, 3)
            rot_coords2 = SL2.coords[:, SL2.side_chain_idx].reshape(-1, 3)
            rsl = cdist(rot_coords1, rot_coords2)
            rsl = rsl.reshape(nrot1, nat1, nrot2, nat2).transpose(0, 2, 1, 3)
            join_rmin = chilife.get_lj_rmin("join_protocol")[()]
            join_eps = chilife.get_lj_eps("join_protocol")[()]

            rmin_ij = join_rmin(SL1.rmin2, SL2.rmin2)
            eps_ij = join_eps(SL1.eps, SL2.eps)

            lj = rmin_ij[None, None, ...] / rsl
            lj = lj * lj * lj
            lj = lj * lj
            lj = lj * lj

            # Cap
            lj[lj > 10] = 10
            # Rep only
            E = eps_ij * (lj * lj)
            E = E.sum(axis=(2, 3))
            weights[-1], _ = reweight_rotamers(E.flatten(), SL1.temp, weights[-1])

    distances = np.concatenate(distances)
    weights = np.concatenate(weights)

    # Calculate distance histogram
    hist, _ = np.histogram(
        distances, weights=weights, range=(min(r), max(r)), bins=len(r)
    )

    # Convolve with normal distribution if non-zero standard deviation is given
    if sigma != 0:
        delta_r = get_delta_r(r)
        _, g = normdist(delta_r, 0, sigma)
        P = np.convolve(hist, g, mode="same")
    else:
        P = hist

    # Normalize distribution
    integral = np.trapz(P, r)
    if integral != 0:
        P /= integral

    return P


def traj_dd(
        SL1: SpinLabelTraj,
        SL2: SpinLabelTraj,
        r: ArrayLike,
        sigma: float,
        **kwargs,
) -> np.ndarray:
    """Calculate a distance distribution from a trajectory of spin labels by calling ``distance_distribution`` on each
    frame and averaging the resulting distributions.

    Parameters
    ----------
    SL1, SL2: SpinLabelTrajectory
        Spin label to use for distance distribution calculation.
    r : ArrayLike
        Distance domain to use when calculating distance distribution.
    sigma : float
        Standard deviation of the gaussian kernel used to smooth the distance distribution
    **kwargs : dict, optional
        Additional keyword arguments to pass to ``distance_distribution`` .

    Returns
    -------
    P: ndarray
        Distance distribution calculated from the provided SpinLabelTrajectories
    """

    # Ensure that the SpinLabelTrajectories have the same number of frames.
    if len(SL1) != len(SL2):
        raise ValueError("SpinLabelTraj objects must have the same length")

    # Calculate the distance distribution for each frame and sum
    P = np.zeros_like(r)
    for _SL1, _SL2 in zip(SL1, SL2):
        P += distance_distribution(_SL1, _SL2, r=r, sigma=sigma, **kwargs)

    # Normalize distance distribution
    P /= np.trapz(P, r)

    return P


def repack(
        protein: Union[mda.Universe, mda.AtomGroup],
        *spin_labels: RotamerEnsemble,
        repetitions: int = 200,
        temp: float = 1,
        energy_func: Callable = get_lj_rep,
        off_rotamer=False,
        **kwargs,
) -> Tuple[mda.Universe, ArrayLike]:
    """Markov chain Monte Carlo repack a protein around any number of SpinLabel or RotamerEnsemble objects.

    Parameters
    ----------
    protein : MDAnalysis.Universe, MDAnalysis.AtomGroup
        Protein to be repacked
    spin_labels : RotamerEnsemble, SpinLabel
        RotamerEnsemble or SpinLabel object placed at site of interest
    repetitions : int
        Number of successful MC samples to perform before terminating the MC sampling loop
    temp : float, ArrayLike
        Temperature (Kelvin) for both clash evaluation and metropolis-hastings acceptance criteria. Accepts a list or
        array like object of temperatures if a temperature schedule is desired
    energy_func : Callable
        Energy function to be used for clash evaluation. Must accept a protein and RotamerEnsemble object and return an
        array of potentials in kcal/mol, with one energy per rotamer in the rotamer ensemble
    off_rotamer : bool
        Boolean argument that decides whether off rotamer sampling is used when repacking the provided residues.
    kwargs : dict
        Additional keyword arguments to be passed to ``mutate`` .

    Returns
    -------
    protein: MDAnalysis.Universe
        MCMC trajectory of local repack
    deltaEs: np.ndarray:
        Change in energy_func score at each accept of MCMC trajectory

    """
    temp = np.atleast_1d(temp)
    KT = {t: GAS_CONST * t for t in temp}

    repack_radius = kwargs.pop("repack_radius") if "repack_radius" in kwargs else None  # Angstroms
    if repack_radius is None:
        repack_radius = max([SL.clash_radius for SL in spin_labels])
    # Construct a new spin labeled protein and preallocate variables to retain monte carlo trajectory
    spin_label_str = " or ".join(
        f"( {spin_label.selstr} )" for spin_label in spin_labels
    )
    protein = mutate(protein, *spin_labels, **kwargs).atoms

    # Determine the residues near the spin label that will be repacked
    repack_residues = protein.select_atoms(f"around {repack_radius} {spin_label_str}").residues

    repack_res_kwargs = spin_labels[0].input_kwargs
    repack_res_kwargs['eval_clash'] = False
    repack_residue_libraries = [
        RotamerEnsemble.from_mda(res, **repack_res_kwargs)
        for res in repack_residues
        if res.resname not in ["GLY", "ALA"]
           and res.resname in chilife.SUPPORTED_RESIDUES
    ]
    repack_residue_libraries += list(spin_labels)

    # Create new labeled protein construct to fill in any missing atoms of repack residues
    protein = mutate(protein, *repack_residue_libraries, **kwargs).atoms

    repack_residues = protein.select_atoms(f"around {repack_radius} {spin_label_str}").residues

    repack_residue_libraries = [
        RotamerEnsemble.from_mda(res, **repack_res_kwargs)
        for res in repack_residues
        if res.resname not in ["GLY", "ALA"]
           and res.resname in chilife.SUPPORTED_RESIDUES
    ]
    repack_residue_libraries += [RotamerEnsemble(RL.res, RL.site, protein, rotlib=RL._rotlib, **repack_res_kwargs)
                                    for RL in spin_labels]

    traj = np.empty((repetitions, *protein.positions.shape))
    deltaEs = []

    sample_freq = np.array(
        [len(res.dihedral_atoms) for res in repack_residue_libraries], dtype=np.float64
    )
    mean = sample_freq[sample_freq > 1].mean()
    sample_freq[sample_freq == 1] = mean
    sample_freq /= sample_freq.sum()

    count = 0
    acount = 0
    bcount = 0
    bidx = 0
    schedule = repetitions / (len(temp) + 1)
    with tqdm(total=repetitions) as pbar:
        while count < repetitions:

            # Randomly select a residue from the repacked residues
            SiteLibrary = repack_residue_libraries[
                np.random.choice(len(repack_residue_libraries), p=sample_freq)
            ]
            if not hasattr(SiteLibrary, "dummy_label"):
                SiteLibrary.dummy_label = SiteLibrary.copy()
                SiteLibrary.dummy_label.protein = protein
                SiteLibrary.dummy_label.mask = np.isin(
                    protein.ix, SiteLibrary.clash_ignore_idx
                )
                SiteLibrary.dummy_label._coords = np.atleast_3d(
                    [protein.atoms[SiteLibrary.dummy_label.mask].positions]
                )

                with np.errstate(divide="ignore"):
                    SiteLibrary.dummy_label.E0 = energy_func(SiteLibrary.dummy_label) - \
                                                 KT[temp[bidx]] * np.log(SiteLibrary.current_weight)

            DummyLabel = SiteLibrary.dummy_label

            coords, weight = SiteLibrary.sample(off_rotamer=off_rotamer)

            with np.errstate(divide="ignore"):
                DummyLabel._coords = np.atleast_3d([coords])
                E1 = energy_func(DummyLabel) - KT[temp[bidx]] * np.log(weight)

            deltaE = E1 - DummyLabel.E0
            deltaE = np.maximum(deltaE, -10.0)

            acount += 1
            # Metropolis-Hastings criteria

            if (
                    E1 < DummyLabel.E0
                    or np.exp(-deltaE / KT[temp[bidx]]) > np.random.rand()
            ):

                deltaEs.append(deltaE)
                try:
                    protein.atoms[DummyLabel.mask].positions = coords
                    DummyLabel.E0 = E1

                except ValueError as err:
                    print(SiteLibrary.name)
                    print(SiteLibrary.atom_names)

                    raise ValueError(err)

                traj[count] = protein.atoms.positions
                SiteLibrary.update_weight(weight)
                count += 1
                bcount += 1
                pbar.update(1)
                if bcount > schedule:
                    bcount = 0
                    bidx = np.minimum(bidx + 1, len(temp) - 1)
            else:
                continue

    logging.info(f"Total counts: {acount}")

    # Load MCMC trajectory into universe.
    if isinstance(protein, (mda.Universe, mda.AtomGroup)):
        protein.universe.load_new(traj)
    else:
        protein.protein.trajectory = chilife.Trajectory(traj, protein)

    return protein, np.squeeze(deltaEs)


def create_library(
        libname: str,
        pdb: str,
        dihedral_atoms: List[List[str]] = [],
        site: int = 1,
        resname: str = None,
        dihedrals: ArrayLike = None,
        weights: ArrayLike = None,
        sigmas: ArrayLike = None,
        permanent: bool = False,
        default: bool = False,
        force: bool = False,
        spin_atoms: List[str] = None,
        description: str = None,
        comment: str = None,
        reference: str = None
) -> None:
    """Add a user defined SpinLabel from a pdb file.

    Parameters
    ----------
    libname : str
        Name for the new rotamer library.
    pdb : str
        Name of (and path to) pdb file containing the user defined spin label structure. This pdb file should contain
        only the desired spin label and no additional residues.
    site : int
        The residue number of the side chain in the pdb file you would like to add.
    resname : str
        3-letter residue code.
    dihedral_atoms : list
        a list of rotatable dihedrals. List should contain sub-lists of 4 atom names. Atom names must be the same as
        defined in the pdb file eg:
        
        .. code-block:: python

            [['CA', 'CB', 'SG', 'SD'],
            ['CB', 'SG', 'SD', 'CE']...]

    dihedrals : ArrayLike, optional
        Array of dihedral angles. If provided the new label object will be stored as a rotamer library with the
        dihedrals provided. Array should be n x m where n is the number of rotamers and m is the number of dihedrals.
    weights : ArrayLike, optional
        Weights associated with the dihedral angles provided by the ``dihedrals`` keyword argument. There should be one
        weight per rotamer and the rotamer weights should sum to 1.
    sigmas : ArrayLike, optional
        Sigma parameter for distributions of dihedral angles. Should be n x m matrix where n is the number of rotamers
        and m is the number of dihedrals. This feature will be used when performing off rotamer samplings.
    permanent: bool
        If set to True the library will be stored in the chilife user_rotlibs directory in addition to the current
        working directory.
    default : bool
        If set to true and permanent is also set to true then this rotamer library will become the default rotamer
        library for the given resname
    force: bool = False,
        If set to True and permanent is also set to true this library will overwrite any existing library with the same
        name.
    spin_atoms : list
        A list of atom names on which the spin density is localized.
    description : str
        A short (< 60 characters) description of the side chain library being created
    comment : str
        Additional information about the rotamer library that may not fit in description.
    reference : str
        Any relevant citations associated with the rotamer library.
    """
    resname = libname[:3] if resname is None else resname
    struct, spin_atoms = pre_add_library(pdb, spin_atoms)
    resi_selection = struct.select_atoms(f"resnum {site}")
    bonds = resi_selection.intra_bonds.indices

    # Convert loaded rotamer ensemble to internal coords
    internal_coords = [
        chilife.get_internal_coords(
            resi_selection,
            preferred_dihedrals=dihedral_atoms,
            bonds=bonds
        )
        for _ in struct.trajectory
    ]

    # set residue number to 1 and remove chain operators so all rotamers are in the ic coordinate frame
    for ic in internal_coords:
        ic.shift_resnum(-(site - 1))
        ic.chain_operators = None
        if len(ic.chains) > 1:
            raise ValueError('The PDB of the label supplied appears to have a chain break. Please check your PDB and '
                             'make sure there are no chain breaks in the desired label and that there are no other '
                             'chains in the pdb file. If the error persists, check to be sure all atoms are the correct '
                             'element as chilife uses the elements to determine if atoms are bonded.')

    # If multi-state pdb extract dihedrals from pdb
    if dihedrals is None:
        dihedrals = np.rad2deg(
            [ic.get_dihedral(1, dihedral_atoms) for ic in internal_coords]
        )

    if dihedrals.shape == (len(dihedrals),):
        dihedrals.shape = (len(dihedrals), 1)

    if weights is None:
        weights = np.ones(len(dihedrals))
        weights /= weights.sum()

    save_dict = prep_restype_savedict(libname, resname, internal_coords,
                                      weights, dihedrals, dihedral_atoms,
                                      sigmas=sigmas, spin_atoms=spin_atoms,
                                      description=description, comment=comment,
                                      reference=reference)

    # Save rotamer library
    np.savez(Path().cwd() / f'{libname}_rotlib.npz', **save_dict, allow_pickle=True)

    if permanent:
        add_library(f'{libname}_rotlib.npz', libname=libname, default=default, force=force)


def create_dlibrary(
        libname: str,
        pdb: str,
        sites: Tuple,
        dihedral_atoms: List[List[List[str]]],
        resname: str = None,
        dihedrals: ArrayLike = None,
        weights: ArrayLike = None,
        permanent: bool = False,
        default: bool = False,
        force: bool = False,
        spin_atoms: List[str] = None,
        description: str = None,
        comment: str = None,
        reference: str = None
) -> None:
    """Add a user defined dSpinLabel from a pdb file.

    Parameters
    ----------
    libname: str,
        Name for the user defined label.
    pdb : str
        Name of (and path to) pdb file containing the user defined spin label structure. This pdb file should contain
        only the desired spin label and no additional residues.
    sites : Tuple[int]
        The residue numbers that the bifunctional label is attached to.
    dihedral_atoms: List[List[List[str]]]
        A list defining the mobile dihedral atoms of both halves of the bifunctional label. There should be  two
        entries, one for each site that the bifunctional label attaches to. Each entry should contain a sublist of
        dihedral definitions. Dihedral definitions are a list of four strings that are the names of the four atoms
        defining the dihedral angle.

            .. code-block:: python

                [
                # Side chain 1
                [['CA', 'CB', 'SG', 'SD'],
                ['CB', 'SG', 'SD', 'CE']...],

                # Side chain 2
                [['CA', 'CB', 'SG', 'SD'],
                ['CB', 'SG', 'SD', 'CE']...]
                ]

    resname : str
        Residue type 3-letter code.
    dihedrals : ArrayLike, optional
        Array of dihedral angles. If provided the new label object will be stored as a rotamer library with the
        dihedrals provided.
    weights : ArrayLike, optional
        Weights associated with the dihedral angles provided by the ``dihedrals`` keyword argument
        permanent: bool
        If set to True the library will be stored in the chilife user_rotlibs directory in addition to the current
        working directory.
    permanent : bool
        If set to True, a copy of the rotamer library will be stored in the chiLife rotamer libraries directory
    default : bool
        If set to true and permanent is also set to true then this rotamer library will become the default rotamer
        library for the given resname
    force: bool = False,
        If set to True and permanent is also set to true this library will overwrite any existing library with the same
        name.
    spin_atoms : list, dict
        List of atom names on which the spin density is localized, e.g ['N', 'O'], or dictionary with spin atom
        names (key 'spin_atoms') and spin atom weights (key 'spin_weights').
    description : str
        A short (< 60 characters) description of the side chain library being created
    comment : str
        Additional information about the rotamer library that may not fit in description.
    reference : str
        Any relevant citations associated with the rotamer library.
    """

    site1, site2 = sites
    increment = site2 - site1
    resname = libname[:3] if resname is None else resname

    if len(dihedral_atoms) != 2:
        dihedral_error = True
    elif not isinstance(dihedral_atoms[0], List):
        dihedral_error = True
    elif len(dihedral_atoms[0][0]) != 4:
        dihedral_error = True
    else:
        dihedral_error = False

    if dihedral_error:
        raise ValueError(
            "dihedral_atoms must be a list of lists where each sublist contains the list of dihedral atoms"
            "for the i and i+{increment} side chains. Sublists can contain any amount of dihedrals but "
            "each dihedral should be defined by exactly four unique atom names that belong to the same "
            "residue number"
        )

    struct, spin_atoms = pre_add_library(pdb, spin_atoms, uniform_topology=False)
    res1 = struct.select_atoms(f'resnum {site1}')
    res2 = struct.select_atoms(f'resnum {site2}')

    # Identify the cap based off of the user defined mobile dihedrals
    nodes = []
    for i, res in enumerate((res1, res2)):
        site = res.resnums[0]
        dh_atoms = [dihedral[1] for dihedral in dihedral_atoms[i]]
        terminal_atom_idx = max(struct.select_atoms(f'resnum {site} and name {" ".join(dh_atoms)}').ix)
        nodes.append(terminal_atom_idx)

    G = nx.DiGraph(struct.bonds.indices.tolist())
    linker = list({node for path in nx.all_simple_paths(G, *nodes) for node in path if node not in nodes})
    G.remove_nodes_from(nodes)

    cap_idxs = set()
    for node in linker:
        cap_idxs |= {a for a in nx.dfs_postorder_nodes(G, node)}

    cap_idxs = list(sorted(cap_idxs))
    ovlp_selection = struct.atoms[cap_idxs]
    csts = ovlp_selection.names
    csts = csts.astype('U4')

    ovlp_selection.residues.resnums = site1
    ovlp_selection.residues.resids = site1
    res1 += ovlp_selection
    res1_bonds = res1.intra_bonds.indices

    IC1 = [chilife.get_internal_coords(res1, dihedral_atoms[0], res1_bonds)
           for ts in struct.trajectory]

    ovlp_selection.residues.resnums = site2
    ovlp_selection.residues.resids = site2
    res2 += ovlp_selection
    res2_bonds = res2.intra_bonds.indices

    IC2 = [chilife.get_internal_coords(res2, dihedral_atoms[1], res2_bonds, cap=cap_idxs)
           for ts in struct.trajectory]

    for ic1, ic2 in zip(IC1, IC2):
        ic1.shift_resnum(-(site1 - 1))
        ic1.chain_operators = None
        ic2.shift_resnum(-(site2 - 1))
        ic2.chain_operators = None
        if len(ic1.chains) > 1 or len(ic2.chains) > 1:
            raise ValueError('The PDB of the label supplied appears to have a chain break. Please check your PDB and '
                             'make sure there are no chain breaks in the desired label and that there are no other '
                             'chains in the pdb file. If the error persists, check to be sure all atoms are the correct '
                             'element as chilife uses the elements to determine if atoms are bonded.')

    # If multi-state pdb extract rotamers from pdb
    if dihedrals is None:
        dihedrals = []
        for IC, resnum, dihedral_set in zip([IC1, IC2], [site1, site2], dihedral_atoms):
            dihedrals.append([[ICi.get_dihedral(1, ddef) for ddef in dihedral_set] for ICi in IC])

    if weights is None:
        weights = np.ones(len(IC1))

    weights /= weights.sum()
    libname = libname + f'ip{increment}'
    save_dict_1 = prep_restype_savedict(libname + 'A', resname, IC1,
                                        weights, dihedrals[0], dihedral_atoms[0],
                                        spin_atoms=spin_atoms, description=description,
                                        comment=comment, reference=reference)
    save_dict_2 = prep_restype_savedict(libname + 'B', resname, IC2,
                                        weights, dihedrals[1], dihedral_atoms[1],
                                        resi=1 + increment,
                                        spin_atoms=spin_atoms)

    # Save individual data sets and zip
    cwd = Path().cwd()
    np.savez(cwd / f'{libname}A_rotlib.npz', **save_dict_1, allow_pickle=True)
    np.savez(cwd / f'{libname}B_rotlib.npz', **save_dict_2, allow_pickle=True)
    np.save(cwd / f'{libname}_csts.npy', csts)

    with zipfile.ZipFile(f'{libname}_drotlib.zip', mode='w') as archive:
        archive.write(f'{libname}A_rotlib.npz')
        archive.write(f'{libname}B_rotlib.npz')
        archive.write(f'{libname}_csts.npy')

    # Cleanup intermediate files
    os.remove(f'{libname}A_rotlib.npz')
    os.remove(f'{libname}B_rotlib.npz')
    os.remove(f'{libname}_csts.npy')

    if permanent:
        add_library(f'{libname}_drotlib.zip', libname=libname, default=default, force=force)


def pre_add_library(
        pdb: str,
        spin_atoms: List[str],
        uniform_topology: bool = True,
) -> Tuple[MDAnalysis.Universe, Dict]:
    """Helper function to sort PDBs, save spin atoms, update lists, etc. when adding a SpinLabel or dSpinLabel.

    Parameters
    ----------
    pdb : str
        Name (and path) of the pdb containing the new label.
    spin_atoms : List[str]
        Atoms of the SpinLabel where the unpaired electron is located.
    uniform_topology : bool
        Assume all rotamers of the library have the same topology (i.e. no differences in atom bonding). If false
        chilife will attempt to find the minimal topology shared between all rotamers for defining internal coordinates.

    Returns
    -------
    struct : MDAnalysis.Universe
        MDAnalysis Universe object containing the rotamer ensemble with each rotamer as a frame. All atoms should be
        properly sorted for consistent construction of internal coordinates.
    spin_atoms : dict
        Dictionary of spin atoms and weights if specified.
    """
    # Sort the PDB for optimal dihedral definitions
    pdb_lines, bonds = sort_pdb(pdb, uniform_topology=uniform_topology, return_bonds=True)
    bonds = get_min_topol(pdb_lines, forced_bonds=bonds)

    # Write a temporary file with the sorted atoms
    if isinstance(pdb_lines[0], list):
        with tempfile.NamedTemporaryFile(suffix=".pdb", mode="w+", delete=False) as tmpfile:
            for i, model in enumerate(pdb_lines):
                tmpfile.write(f"MODEL {i + 1}\n")
                for atom in model:
                    tmpfile.write(atom)
                tmpfile.write("ENDMDL\n")
    else:
        with tempfile.NamedTemporaryFile(suffix=".pdb", mode="w+", delete=False) as tmpfile:
            for line in pdb_lines:
                tmpfile.write(line)

    # Load sorted atom pdb using MDAnalysis and remove tempfile
    struct = mda.Universe(tmpfile.name, in_memory=True)
    struct.universe.add_bonds(bonds)
    os.remove(tmpfile.name)

    # Store spin atoms if provided
    if spin_atoms is not None:
        if isinstance(spin_atoms, str):
            spin_atoms = spin_atoms.split()

        if isinstance(spin_atoms, dict):
            spin_weights = list(spin_atoms.values())
            spin_atoms = list(spin_atoms.keys())

        else:
            w = 1 / len(spin_atoms)
            spin_weights = [w for _ in spin_atoms]

        spin_atoms = {'spin_atoms': spin_atoms, 'spin_weights': spin_weights}

    return struct, spin_atoms


def prep_restype_savedict(
        libname: str,
        resname: str,
        internal_coords: List[ProteinIC],
        weights: ArrayLike,
        dihedrals: ArrayLike,
        dihedral_atoms: ArrayLike,
        sigmas: ArrayLike = None,
        resi: int = 1,
        spin_atoms: List[str] = None,
        description: str = None,
        comment: str = None,
        reference: str = None
) -> Dict:
    """Helper function to add new residue types to chilife

    Parameters
    ----------
    libname : str
        Name of residue to be stored.
    resname : str
        Residue name (3-letter code)
    internal_coords : List[ProteinIC]
        list of internal coordinates of the new residue type.
    weights : ArrayLike
        Array of weights corresponding to each rotamer of the library
    dihedrals : ArrayLike
        Array of mobile dihedral angles for each rotamer fo the library
    dihedral_atoms : ArrayLike
        Definition of mobile dihedral angles for a single structure. Should be a list of 4 string lists where the
        four strings are the names of the atoms that define the dihedral.
    sigmas : ArrayLike
        Array of sigma values for each dihedral of each rotamer.
    resi: int
        The residue number to be stored.
    spin_atoms: List[str]
        A list of atom names corresponding to the atoms where the spin density resides.
    description : str
        A short (< 60 characters) description of the side chain library being created
    comment : str
        Additional information about the rotamer library that may not fit in description.
    reference : str
        Any relevant citations associated with the rotamer library.
    Returns
    -------
    save_dict : dict
        Dictionary of all the data needed to build a RotamerEnsemble.
    """
    # Extract coordinates and transform to the local frame
    bb_atom_idx = [
        i for i, atom in enumerate(internal_coords[0].atoms) if atom.name in ["N", "CA", "C"]
    ]
    coords = internal_coords[0].coords.copy()
    ori, mx = local_mx(*coords[bb_atom_idx])
    coords = (coords - ori) @ mx

    if len(internal_coords) > 1:
        coords = np.array([(IC.coords - ori) @ mx for IC in internal_coords])
    elif len(dihedrals) > 1:
        coords = np.array([internal_coords.set_dihedral(dihe, resi, dihedral_atoms) for dihe in dihedrals])
    else:
        if coords.ndim == 2:
            coords = np.expand_dims(coords, axis=0)

    atom_types = np.array([atom.atype for atom in internal_coords[0].atoms])
    atom_names = np.array([atom.name for atom in internal_coords[0].atoms])

    if np.any(np.isnan(coords)):
        idxs = np.argwhere(np.isnan(np.sum(coords, axis=(1, 2)))).T[0]
        adxs = np.argwhere(np.isnan(np.sum(coords, axis=(0, 2)))).T[0]
        adxs = atom_names[adxs]

        print(internal_coords[0].atom_dict['dihedrals'][1][(5, ('O3', 'Cu1', 'NE2'))])
        print(internal_coords[0].zmats[1].shape)
        print(internal_coords[0].get_dihedral(5, ['NE2', 'Cu1', 'O3', 'C11']))
        print(internal_coords[4].coords)

        raise ValueError(
            f'Coords of rotamer {" ".join((str(idx) for idx in idxs))} at atoms {" ".join((str(idx) for idx in adxs))} '
            f'cannot be converted to internal coords')

    save_dict = {'rotlib': libname,
                 'resname': resname,
                 'coords': coords,
                 'internal_coords': internal_coords,
                 'weights': weights,
                 'atom_types': atom_types,
                 'atom_names': atom_names,
                 'dihedrals': dihedrals,
                 'dihedral_atoms': dihedral_atoms,
                 'description': description,
                 'comment': comment,
                 'reference': reference}

    if sigmas is None:
        pass
    elif sigmas.shape == dihedrals.shape:
        save_dict['sigmas'] = sigmas
    elif sigmas.shape == (*dihedrals.shape, 3):
        save_dict['sigmas'] = sigmas[..., 2]
        save_dict['locs'] = sigmas[..., 1]
        save_dict['skews'] = sigmas[..., 0]

    if spin_atoms:
        save_dict.update(spin_atoms)

    save_dict['type'] = 'chilife rotamer library'
    save_dict['format_version'] = 1.1

    return save_dict


def add_library(filename: Union[str, Path], libname: str = None, default: bool = False, force: bool = False):
    """
    Add the provided rotamer library to the chilife rotamer library directory so that it does not need to be
    in the working directory when utilizing.

    Parameters
    ----------
    filename : str, Path
        Name or Path object oof the rotamer library npz file.
    libname : str
        Unique name of the rotamer library
    default : bool
        If True, chilife will make this the default rotamer library for this residue type.
    force : bool
        If True, chilife will overwrite any existing rotamer library with the same name if it exists.
    """

    store_loc = (RL_DIR / f"user_rotlibs/") / filename
    filename = Path(filename)
    if libname is None:
        libname = re.sub("_d{0,1}rotlib.(npz|zip)", "", filename.name)

    library = chilife.read_library(Path(filename), None, None)
    drotlib = False
    if isinstance(library, tuple):
        drotlib = True
        library, _, _ = library

    resname = str(library['resname'])
    add_to_defaults(resname, libname, default)
    if force or not store_loc.exists():
        shutil.copy(filename, str(store_loc))
        if drotlib:
            global USER_dLIBRARIES
            USER_dLIBRARIES.add(libname)
        else:
            global USER_LIBRARIES
            USER_LIBRARIES.add(libname)
        add_dihedral_def(libname, library['dihedral_atoms'].tolist(), force=force)
    else:
        raise NameError("A rotamer library with this name already exists! Please choose a different name or do"
                        "not store as a permanent rotamer library")


def add_dihedral_def(name: str, dihedrals: ArrayLike, force: bool = False) -> None:
    """Helper function to add the dihedral definitions of user defined labels and libraries to the chilife knowledge
    base.

    Parameters
    ----------
    name : str
        Name of the residue.
    dihedrals : ArrayLike
        List of lists of atom names defining the dihedrals.
    force : bool
        Overwrite any dihedral definition with the same name if it exists.
    """

    # Reload in case there were other changes
    if not add_to_toml(DATA_DIR / "dihedral_defs.toml", key=name, value=dihedrals, force=force):
        raise ValueError(f'There is already a dihedral definition for {name}. Please choose a different name.')

    # Add to active dihedral def dict
    chilife.dihedral_defs[name] = dihedrals


def remove_library(name: str, prompt: bool = True):
    """
    Removes a library from the chilife rotamer library directory and from the chilife dihedral definitions

    Parameters
    ----------
    name : str
        Name of the rotamer library
    prompt : bool
        If set to False al warnings will be ignored.
    """
    global USER_LIBRARIES, USER_dLIBRARIES
    if (name not in USER_LIBRARIES) and (name not in USER_dLIBRARIES) and prompt:
        raise ValueError(f'{name} is not in the set of user labels or user dLables. Check to make sure you have the '
                         f'right label. Note that only user labels can be removed.')

    if prompt:
        ans = input(f'WARNING: You have requested the permanent removal of the {name} label/rotamer library. \n'
                    f'Are you sure you want to do this? (y/n)')
        if ans.lower().startswith('y'):
            pass
        elif ans.lower().startswith('n'):
            print('Canceling label removal')
            return None
        else:
            print(f'"{ans}" is not an intelligible answer. Canceling label removal')
            return None

    # Remove files
    files = list((RL_DIR / 'user_rotlibs').glob(f'{name}*')) + \
            list((RL_DIR / 'residue_internal_coords').glob(f'{name}*')) + \
            list((RL_DIR / 'residue_pdbs').glob(f'{name}*'))

    for file in files:
        if file.exists():
            os.remove(str(file))

    if name in USER_dLIBRARIES:
        USER_dLIBRARIES.remove(name)

    if name in USER_LIBRARIES:
        USER_LIBRARIES.remove(name)

    if name in dihedral_defs:
        del dihedral_defs[name]

    remove_from_toml(DATA_DIR / 'dihedral_defs.toml', name)
    remove_from_defaults(name)


def add_to_toml(file: Union[str, Path], key: str, value: Union[str, List, dict], force: bool = False):
    """Helper function to add new key:value pairs to toml files like dihedral_defs.toml

    Parameters
    ----------
    file : str, Path
        Name or path to the toml file to be edited.
    key : str
        toml entry to be added.
    value : str, List, dict
        Item to add to the toml file under the key.
    force : bool
        If True, any existing data under the `key` entry will be overwritten.
    """
    with open(file, 'r') as f:
        local = rtoml.load(f)

    if key in local and not force:
        return False

    if key is not None:
        local[key] = value

    with open(file, 'w') as f:
        rtoml.dump(local, f)

    return True


def remove_from_toml(file: Union[str, Path], entry: str):
    """
    Remove an entry for a chilife toml file
    Parameters
    ----------
    file : str, Path
        Name or Path of the toml file to be edited.
    entry : str
        Entry or key of the toml file to be removed.
    """
    with open(file, 'r') as f:
        local = rtoml.load(f)

    if entry in local:
        del local[entry]

    with open(file, 'w') as f:
        rtoml.dump(local, f)

    return True


def add_to_defaults(resname: str, rotlib_name: str, default: bool = False):
    """
    Helper function to add a rotamer library to the defaults stack.

    Parameters
    ----------
    resname : str
        3 letter code name of the residue.
    rotlib_name : str
        Name of the rotamer library.
    default : bool
        If True the rotamer library will be added to the top of the defaults stacking making it the new default
        rotamer library for the residue type. If False it will be added to the bottom of the stack and will only be the
        default if there are no other rotamer libraries for the residue type.
    """
    file = RL_DIR / 'defaults.toml'
    with open(file, 'r') as f:
        local = rtoml.load(f)
        backup = deepcopy(local)

    if resname not in local:
        local[resname] = []
    if resname not in chilife.rotlib_defaults:
        chilife.rotlib_defaults[resname] = []

    pos = 0 if default else len(local[resname])

    if rotlib_name not in chilife.rotlib_defaults[resname]:
        chilife.rotlib_defaults[resname].insert(pos, rotlib_name)
    if rotlib_name not in local[resname]:
        local[resname].insert(pos, rotlib_name)

    safe_save(file, local, backup)


def remove_from_defaults(rotlib_name: str):
    """
    Helper function to remove a rotamer library from the defaults stack.
    Parameters
    ----------
    rotlib_name : str
        Name of the rotamer library to be removed.
    """
    file = RL_DIR / 'defaults.toml'
    with open(file, 'r') as f:
        local = rtoml.load(f)
        backup = deepcopy(local)

    keys = [key for key, val in local.items() if rotlib_name in val]

    for key in keys:
        chilife.rotlib_defaults[key].remove(rotlib_name)
        local[key].remove(rotlib_name)
        if local[key] == []:
            del local[key]

    safe_save(file, local, backup)


def safe_save(file: Union[str, Path], data: dict, backup: dict):
    """
    Helper function to save toml files. Will preserve original toml file if there is an error.

    Parameters
    ----------
    file : str, Path
        Name of the toml file to be saved.
    data : dict
        Data to save in the toml file.
    backup : dict
        Original data from the toml file to save if there is an error.
    """
    try:
        with open(file, 'w') as f:
            rtoml.dump(data, f)
    except rtoml._rtoml.TomlSerializationError:
        with open(file, 'w') as f:
            rtoml.dump(backup, f)
        raise


def _print_monofunc(files):
    print(f"{'monofunctional' + ':':^60}")
    print("-" * 60)

    for file in files:
        with np.load(file, allow_pickle=True) as f:
            for key in f.keys():
                resname = f['resname']
                if 'description' in f:
                    descr = f['description']
                else:
                    descr = "No description"

        print(f'{resname:<8} : {descr}')
    print("-" * 60)


def _print_bifunc(files):
    print(f"{'bifunctional' + ':':^60}")
    print("-" * 60)

    for file in files:
        with zipfile.ZipFile(file, 'r') as archive:
            name = archive.namelist()[0]
            resname = name[:3]
            context = name[3:6]
            with archive.open(name) as of:
                with np.load(of, allow_pickle=True) as f:
                    if 'description' in f:
                        descr = f['description']
                    else:
                        descr = "No description"

        print(f'{resname:<4}{context:<4} : {descr}')
    print("-" * 60)


def list_available_rotlibs():
    """
    Lists residue types and rotamer libraries that are currently available. More information on any individual rotamer
    library by using the rotlib_info function.
    """

    print()

    dirs = {'current working directory': Path.cwd()}
    dirs.update({f'user rotlib directory {n + 1}': u_dir for n, u_dir in enumerate(USER_RL_DIR)})
    dirs.update({'chilife rotlib directory': RL_DIR / 'user_rotlibs'})

    for dname, dpath in dirs.items():
        monofunctional = tuple(dpath.glob('*_rotlib.npz'))
        bifunctional = tuple(dpath.glob('*_drotlib.zip'))
        if len(monofunctional) + len(bifunctional) > 0:

            print("*" * 60)
            print(f"*{f'Rotlibs in {dname}':^58}*")
            if 'user' in dname:
                print(f'*{dname.name:^58}*')
            print("*" * 60)

            if len(monofunctional) > 0 : _print_monofunc(monofunctional)
            if len(bifunctional) > 0 : _print_bifunc(bifunctional)

    print()
    print("*" * 60)
    print(f"*{'DUNBRACK ROTLIBS':^58}*")
    print(f"*{'ARG, ASN, ASP, CSY, GLN, GLU, HIS, ILE, LEU, LYS, MET,':^58}*")
    print(f"*{'PHE, PRO, SER, THR, TRP, TYR, VAL':^58}*")
    print(f"*{'(no ALA, GLY}':^58}*")
    print("*" * 60)


def add_rotlib_dir(directory: Union[Path, str]) -> None:
    """Add a directory to search for rotlibs when none are found in te working directory. This directory will be
    searched before the chiLife default rotlib directory."""
    directory = Path(directory)
    with open(RL_DIR / 'additional_rotlib_dirs.txt', 'a+') as f:
        f.write(str(directory))
        f.write('\n')

    USER_RL_DIR.append(Path(directory))


def remove_rotlib_dir(directory: Union[Path, str]) -> None:
    directory = Path(directory)
    if directory in USER_RL_DIR:
        USER_RL_DIR.remove(directory)

    with open(RL_DIR / 'additional_rotlib_dirs.txt', 'r') as f:
        TMP = [Path(line.strip()) for line in f.readlines() if Path(line.strip()) != directory]

    with open(RL_DIR / 'additional_rotlib_dirs.txt', 'w') as f:
        for p in TMP:
            f.write(str(p))
            f.write('\n')


def get_possible_rotlibs(rotlib: str,
                         suffix: str,
                         extension: str,
                         return_all: bool = False,
                         was_none: bool = False) -> Union[Path, None]:
    """

    """
    cwd = Path.cwd()
    sufplusex = '_' + suffix + extension
    # Assemble a list of possible rotlib paths starting in the current directory
    possible_rotlibs = [Path(rotlib),
                        cwd / rotlib,
                        cwd / (rotlib + extension),
                        cwd / (rotlib + sufplusex)]

    possible_rotlibs += list(cwd.glob(f'{rotlib}*{sufplusex}'))
    # Then in the user defined rotamer library directory
    for pth in chilife.USER_RL_DIR:
        possible_rotlibs += list(pth.glob(f'{rotlib}*{sufplusex}'))

    if not was_none:
        possible_rotlibs += list((chilife.RL_DIR / 'user_rotlibs').glob(f'*{rotlib}*'))

    if return_all:
        rotlib = []
    for possible_file in possible_rotlibs:
        if possible_file.exists() and return_all:
                rotlib.append(possible_file)
        elif possible_file.exists():
            rotlib = possible_file
            break
    else:
        if isinstance(rotlib, str) and was_none and rotlib in chilife.rotlib_defaults:
            rotlib = chilife.RL_DIR / 'user_rotlibs' / (chilife.rotlib_defaults[rotlib][0] + sufplusex)

        elif not isinstance(rotlib, list) or rotlib == []:
            rotlib = None

    # rotlib lists need to be sorted to prevent position mismatches for results with tests.
    if isinstance(rotlib, list):
        rotlib = sorted(rotlib)

    return rotlib


def rotlib_info(rotlib: str):
    """
    Display detailed information about the rotamer library.

    Parameters
    ----------
    rotlib : str
        Name of the rotamer library to print the information of.

    """
    pass



from typing import Tuple, Dict, Union, BinaryIO, TextIO
from collections import defaultdict
from hashlib import sha256
from pathlib import Path
import pickle
import shutil
from io import StringIO, BytesIO
import zipfile

import numpy as np
from scipy.stats import gaussian_kde
from memoization import cached, suppress_warnings
import MDAnalysis as mda

import chilife
from .RotamerEnsemble import RotamerEnsemble
from .SpinLabel import SpinLabel
from .dRotamerEnsemble import dRotamerEnsemble
from .IntrinsicLabel import IntrinsicLabel
from .Protein import MolecularSystem
from .ProteinIC import ProteinIC

#                 ID    name   res  chain resnum      X     Y      Z      q      b              elem
fmt_str = "ATOM  {:5d} {:^4s} {:3s} {:1s}{:4d}    {:8.3f}{:8.3f}{:8.3f}{:6.2f}{:6.2f}          {:>2s}  \n"


def read_distance_distribution(file_name: str) -> Tuple[np.ndarray, np.ndarray]:
    """Reads a DEER distance distribution file in the DeerAnalysis or similar format.

    Parameters
    ----------
    file_name : str
        File name of distance distribution

    Returns
    -------
    r: np.ndarray
        Domain of distance distribution in the same units as the file.
    p: np.ndarray
        Probability density over r normalized such that the integral of p over r is 1.
    """

    # Load DA file
    data = np.loadtxt(file_name)

    # Convert nm to angstroms
    r = data[:, 0] * 10

    # Extract distance domain coordinates
    p = data[:, 1]
    return r, p


def hash_file(file: Union[Path, BinaryIO]):
    hash = sha256()
    with open(file, 'rb') as f:
        while True:
            block = f.read(hash.block_size)
            if not block:
                break
            hash.update(block)

    return hash.hexdigest()


suppress_warnings()
@cached(custom_key_maker=hash_file)
def read_rotlib(rotlib: Union[Path, BinaryIO] = None) -> Dict:
    """Reads RotamerEnsemble for stored spin labels.

    Parameters
    ----------
    rotlib : Path, BinaryIO
        Path object pointing to the rotamer library. Or a BytesIO object containing the rotamer library file
    Returns
    -------
    lib: dict
        Dictionary of SpinLabel rotamer ensemble attributes including coords, weights, dihedrals etc.

    """
    with np.load(rotlib, allow_pickle=True) as files:
        lib = dict(files)

    del lib["allow_pickle"]

    if "sigmas" not in lib:
        lib["sigmas"] = np.array([])

    lib["_rdihedrals"] = np.deg2rad(lib["dihedrals"])
    lib["_rsigmas"] = np.deg2rad(lib["sigmas"])
    lib['rotlib'] = str(lib['rotlib'])
    lib['type'] = str(lib['type'])
    lib['format_version'] = float(lib['format_version'])
    return lib


# @cached(custom_key_maker=hash_file)
def read_drotlib(rotlib: Path) -> Tuple[dict]:
    """Reads RotamerEnsemble for stored spin labels.

        Parameters
        ----------
        rotlib : Path
            Path to the rotamer library file.
        Returns
        -------
        lib: Tuple[dict]
            Dictionaries of rotamer library attributes including sub_residues .

        """

    with zipfile.ZipFile(rotlib, 'r') as archive:
        for f in archive.namelist():
            if 'csts' in f:
                with archive.open(f) as of:
                    csts = np.load(of)
            elif f[-12] == 'A':
                with archive.open(f) as of:
                    libA = read_rotlib.__wrapped__(of)
            elif f[-12] == 'B':
                with archive.open(f) as of:
                    libB = read_rotlib.__wrapped__(of)

    return libA, libB, csts


@cached
def read_bbdep(res: str, Phi: int, Psi: int) -> Dict:
    """Read the Dunbrack rotamer library for the provided residue and backbone conformation.

    Parameters
    ----------
    res : str
        3-letter residue code
    Phi : int
        Protein backbone Phi dihedral angle for the provided residue
    Psi : int
        Protein backbone Psi dihedral angle for the provided residue

    Returns
    -------
    lib: dict
        Dictionary of arrays containing rotamer library information in cartesian and dihedral space
    """
    lib = {}
    Phi, Psi = str(Phi), str(Psi)

    # Read residue internal coordinate structure
    with open(chilife.RL_DIR / f"residue_internal_coords/{res.lower()}_ic.pkl", "rb") as f:
        ICs = pickle.load(f)

    atom_types = ICs.atom_types.copy()
    atom_names = ICs.atom_names.copy()

    maxchi = 5 if res in chilife.SUPPORTED_BB_LABELS else 4
    nchi = np.minimum(len(chilife.dihedral_defs[res]), maxchi)

    if res not in ("ALA", "GLY"):
        library = "R1C.lib" if res in chilife.SUPPORTED_BB_LABELS else "ALL.bbdep.rotamers.lib"
        start, length = chilife.rotlib_indexes[f"{res}  {Phi:>4}{Psi:>5}"]

        with open(chilife.RL_DIR / library, "rb") as f:
            f.seek(start)
            rotlib_string = f.read(length).decode()
            s = StringIO(rotlib_string)
            s.seek(0)
            data = np.genfromtxt(s, usecols=range(maxchi + 4, maxchi + 5 + 2 * maxchi))

        lib["weights"] = data[:, 0]
        lib["dihedrals"] = data[:, 1: nchi + 1]
        lib["sigmas"] = data[:, maxchi + 1: maxchi + nchi + 1]
        dihedral_atoms = chilife.dihedral_defs[res][:nchi]

        # Calculate cartesian coordinates for each rotamer
        coords = []
        internal_coords = []
        for r in lib["dihedrals"]:
            ICn = ICs.copy().set_dihedral(np.deg2rad(r), 1, atom_list=dihedral_atoms)

            coords.append(ICn.to_cartesian())
            internal_coords.append(ICn)

    else:
        lib["weights"] = np.array([1])
        lib["dihedrals"], lib["sigmas"], dihedral_atoms = [], [], []
        coords = [ICs.to_cartesian()]
        internal_coords = [ICs.copy()]

    # Get origin and rotation matrix of local frame
    mask = np.in1d(atom_names, ["N", "CA", "C"])
    ori, mx = chilife.local_mx(*coords[0][mask])

    # Set coords in local frame and prepare output
    lib["coords"] = np.array([(coord - ori) @ mx for coord in coords])
    lib["internal_coords"] = internal_coords
    lib["atom_types"] = np.asarray(atom_types)
    lib["atom_names"] = np.asarray(atom_names)
    lib["dihedral_atoms"] = np.asarray(dihedral_atoms)
    lib["_rdihedrals"] = np.deg2rad(lib["dihedrals"])
    lib["_rsigmas"] = np.deg2rad(lib["sigmas"])
    lib['rotlib'] = res

    # hacky solution to experimental backbone dependent rotlibs
    if res == 'R1C':
        lib['spin_atoms'] = np.array(['N1', 'O1'])
        lib['spin_weights'] = np.array([0.5, 0.5])

    return lib


def read_library(rotlib: str, Phi: float = None, Psi: float = None) -> Dict:
    """Generalized wrapper function to aid selection of rotamer library reading function.

    Parameters
    ----------
    rotlib : str, Path
        3 letter residue code
    Phi : float
        Protein backbone Phi dihedral angle for the provided residue
    Psi : float
        Protein backbone Psi dihedral angle for the provided residue
    Returns
    -------
    lib: dict
        Dictionary of arrays containing rotamer library information in cartesian and dihedral space
    """
    backbone_exists = Phi and Psi

    if backbone_exists:
        Phi = int((Phi // 10) * 10)
        Psi = int((Psi // 10) * 10)
    # Use helix backbone if not otherwise specified
    else:
        Phi, Psi = -60, -50

    if isinstance(rotlib, Path):
        if rotlib.suffix == '.npz':
            return read_rotlib(rotlib)
        elif rotlib.suffix == '.zip':
            return read_drotlib(rotlib)
        else:
            raise ValueError(f'{rotlib.name} is not a valid rotamer library file type')
    elif isinstance(rotlib, str):
        return read_bbdep(rotlib, Phi, Psi)
    else:
        raise ValueError(f'{rotlib} is not a valid rotamer library')


def save(
        file_name: str,
        *molecules: Union[RotamerEnsemble, MolecularSystem, mda.Universe, mda.AtomGroup, str],
        protein_path: Union[str, Path] = None,
        mode: str = 'w',
        **kwargs,
) -> None:
    """Save a pdb file of the provided labels and proteins

    Parameters
    ----------
    file_name : str
        Desired file name for output file. Will be automatically made based off of protein name and labels if not
        provided.
    *molecules : RotmerLibrary, chiLife.Protein, mda.Universe, mda.AtomGroup, str
        Object(s) to save. Includes RotamerEnsemble, SpinLabels, dRotamerEnsembles, proteins, or path to protein pdb.
        Can add as many as desired, except for path, for which only one can be given.
    protein_path : str, Path
        The Path to a pdb file to use as the protein object.
    **kwargs :
        Additional arguments to pass to ``write_labels``

    Returns
    -------
    None
    """

    if isinstance(file_name, tuple(molecule_class.keys())):
        molecules = list(molecules)
        molecules.insert(0, file_name)
        file_name = None

    # Check for filename at the beginning of args
    tmolecules = defaultdict(list)
    for mol in molecules:
        mcls = [val for key, val in molecule_class.items() if isinstance(mol, key)]
        if mcls == []:
            raise TypeError(f'{type(mol)} is not a supported type for this function. '
                            'chiLife can only save objects of the following types:\n'
                            + ''.join(f'{key.__name__}\n' for key in molecule_class) +
                            'Please check that your input is compatible')

        tmolecules[mcls[0]].append(mol)
    molecules = tmolecules

    # Ensure only one protein path was provided (for now)
    protein_path = [] if protein_path is None else [protein_path]
    if len(protein_path) > 1:
        raise ValueError('More than one protein path was provided. C')
    elif len(protein_path) == 1:
        protein_path = protein_path[0]
    else:
        protein_path = None

    # Create a file name from protein information
    if file_name is None:
        if protein_path is not None:
            f = Path(protein_path)
            file_name = f.name[:-4]
        else:
            file_name = ""
            for protein in molecules['molcart']:
                if getattr(protein, "filename", None):
                    file_name += ' ' + Path(protein.filename).name
                    file_name = file_name[:-4]

        if file_name == '':
            file_name = "No_Name_Protein"

        # Add spin label information to file name
        if 0 < len(molecules['rotens']) < 3:
            for rotens in molecules['rotens']:
                file_name += f"_{rotens.name}"
        else:
            file_name += "_many_labels"

        file_name += ".pdb"
        file_name = file_name.strip()

    if protein_path is not None:
        print(protein_path, file_name)
        shutil.copy(protein_path, file_name)
        pdb_file = open(file_name, 'a+')
    else:
        pdb_file = open(file_name, mode)

    for protein in molecules['molcart']:
        write_protein(pdb_file, protein)

    for ic in molecules['molic']:
        write_ic(pdb_file, ic)

    if len(molecules['rotens']) > 0:
        write_labels(pdb_file, *molecules['rotens'], **kwargs)



def write_protein(pdb_file: TextIO, protein: Union[mda.Universe, mda.AtomGroup]) -> None:
    """Helper function to write protein pdbs from mdanalysis objects.

    Parameters
    ----------
    pdb_file : TextIO
        File to save the protein to
    protein : MDAnalysis.Universe, MDAnalysis.AtomGroup
        MDAnalyiss object to save

    Returns
    -------
    None
    """

    # Change chain identifier if longer than 1
    available_segids = iter('ABCDEFGHIJKLMNOPQRSTUVWXYZ')
    for seg in protein.segments:
        if len(seg.segid) > 1:
            seg.segid = next(available_segids)
    if isinstance(protein, (mda.AtomGroup, mda.Universe)):
        traj = protein.universe.trajectory
        name = Path(protein.universe.filename) if protein.universe.filename is not None else Path(pdb_file.name)
        name = name.name
    else:
        traj = protein.trajectory
        name = protein.fname

    if name is None:
        name = Path(pdb_file.name).name

    name = name[:-4] if name.endswith(".pdb") else name

    pdb_file.write(f'HEADER {name}\n')
    for mdl, ts in enumerate(traj):
        pdb_file.write(f"MODEL {mdl}\n")
        [
            pdb_file.write(
                fmt_str.format(
                    atom.index,
                    atom.name,
                    atom.resname[:3],
                    atom.segid,
                    atom.resnum,
                    *atom.position,
                    1.00,
                    1.0,
                    atom.type,
                )
            )
            for atom in protein.atoms
        ]
        pdb_file.write("TER\n")
        pdb_file.write("ENDMDL\n")


def write_ic(pdbfile: TextIO, ic: chilife.ProteinIC) -> None:
    """
    Write a :class:`chilife.ProteinIC` internal coordinates object to a pdb file.

    Parameters
    ----------
    pdbfile : TextIO
        open file or io object.
    ic: chilife.ProteinIC
        chiLife internal coordinates object.
    """
    pdbfile.write('MODEL\n')
    for atom, coord in zip(ic.atoms, ic.coords):
        pdbfile.write(fmt_str.format(atom.index + 1, atom.name, atom.resn, 'A', atom.resi,
                                     coord[0], coord[1], coord[2],
                                     1.0, 1.0, atom.atype))
    pdbfile.write('ENDMDL\n')


def write_labels(pdb_file: TextIO, *args: SpinLabel, KDE: bool = True, sorted: bool = True) -> None:
    """Lower level helper function for saving SpinLabels and RotamerEnsembles. Loops over SpinLabel objects and appends
    atoms and electron coordinates to the provided file.

    Parameters
    ----------
    file : str
        File name to write to.
    *args: SpinLabel, RotamerEnsemble
        RotamerEnsemble or SpinLabel objects to be saved.
    KDE: bool
        Perform kernel density estimate smoothing on rotamer weights before saving. Usefull for uniformly weighted
        RotamerEnsembles or RotamerEnsembles with lots of rotamers
    sorted : bool
        Sort rotamers by weight befroe saving.
    Returns
    -------
    None
    """

    # Write spin label models
    for k, label in enumerate(args):
        pdb_file.write(f"HEADER {label.name}\n")

        # Save models in order of weight

        sorted_index = np.argsort(label.weights)[::-1] if sorted else np.arange(len(label.weights))
        norm_weights = label.weights / label.weights.max()
        if isinstance(label, dRotamerEnsemble):
            sites = np.concatenate([np.ones(len(label.RE1.atoms), dtype=int) * int(label.site1),
                                    np.ones(len(label.RE2.atoms), dtype=int) * int(label.site2)])
        else:
            sites = [atom.resi for atom in label.atoms]
        for mdl, (conformer, weight) in enumerate(
                zip(label.coords[sorted_index], norm_weights[sorted_index])
        ):
            pdb_file.write(f"MODEL {mdl}\n")

            [
                pdb_file.write(
                    fmt_str.format(
                        i,
                        label.atom_names[i],
                        label.res[:3],
                        label.chain,
                        sites[i],
                        *conformer[i],
                        weight,
                        1.00,
                        label.atom_types[i],
                    )
                )
                for i in range(len(label.atom_names))
            ]
            pdb_file.write("TER\n")
            pdb_file.write("ENDMDL\n")

    # Write electron density at electron coordinates
    for k, label in enumerate(args):
        if not hasattr(label, "spin_centers"):
            continue
        if not np.any(label.spin_centers):
            continue

        pdb_file.write(f"HEADER {label.name}_density\n")
        spin_centers = np.atleast_2d(label.spin_centers)

        if KDE and len(spin_centers) > 5:
            # Perform gaussian KDE to determine electron density
            gkde = gaussian_kde(spin_centers.T, weights=label.weights)

            # Map KDE density to pseudoatoms
            vals = gkde.pdf(spin_centers.T)

        else:
            vals = label.weights
        norm_weights = vals / vals.max()
        [
            pdb_file.write(
                fmt_str.format(
                    i,
                    "NEN",
                    label.label[:3],
                    label.chain,
                    int(label.site),
                    *spin_centers[i],
                    norm_weights[i],
                    1.00,
                    "N",
                )
            )
            for i in range(len(norm_weights))
        ]

        pdb_file.write("TER\n")

molecule_class = {RotamerEnsemble: 'rotens',
                  dRotamerEnsemble: 'rotens',
                  IntrinsicLabel: 'rotens',
                  mda.Universe: 'molcart',
                  mda.AtomGroup: 'molcart',
                  MolecularSystem: 'molcart',
                  ProteinIC: 'molic'}


rotlib_formats = {1.0: (
    'rotlib',  #
    'resname',
    'coords',
    'internal_coords',
    'weights',
    'atom_types',
    'atom_names',
    'dihedrals',
    'dihedral_atoms',
    'type',
    'format_version'
)}

rotlib_formats[1.1] = *rotlib_formats[1.0], 'description', 'comment', 'reference'





import networkx as nx
from copy import deepcopy
from itertools import combinations
import logging
import warnings
import inspect
from functools import partial
from pathlib import Path

import numpy as np
from scipy.spatial import cKDTree
import scipy.optimize as opt
import MDAnalysis as mda

import chilife


class dRotamerEnsemble:
    backbone_atoms = ["H", "N", "CA", "HA", "C", "O"]

    def __init__(self, res, sites, protein=None, chain=None, rotlib=None, **kwargs):
        """ """
        self.res = res
        self.site1, self.site2 = sorted(sites)
        self.site = self.site1
        self.increment = self.site2 - self.site1
        self.kwargs = kwargs

        self.protein = protein
        self.chain = chain if chain is not None else self.guess_chain()
        self.protein_tree = self.kwargs.setdefault("protein_tree", None)

        self.forgive = kwargs.setdefault("forgive", 0.95)
        self._clash_ori_inp = kwargs.setdefault("clash_ori", "cen")
        self.restraint_weight = kwargs.pop("restraint_weight") if "restraint_weight" in kwargs else 222
        self.torsion_weight = kwargs.pop("torsion_weight") if "torsion_weight" in kwargs else 5
        self.alignment_method = kwargs.setdefault("alignment_method", "bisect".lower())
        self.dihedral_sigmas = kwargs.setdefault("dihedral_sigmas", 25)
        self._exclude_nb_interactions = kwargs.setdefault('exclude_nb_interactions', 3)
        self._minimize = kwargs.pop("minimize", True)
        self.min_method = kwargs.pop('min_method', 'L-BFGS-B')
        self.trim_tol = kwargs.pop('trim_tol', 0.005)
        self._do_trim = kwargs.pop('trim', True)
        self.eval_clash = kwargs.pop("eval_clash", True)
        self.energy_func = kwargs.setdefault("energy_func", chilife.get_lj_energy)
        self.temp = kwargs.setdefault("temp", 298)
        self.get_lib(rotlib)
        self.create_ensembles()

        self.RE1.backbone_to_site()
        self.RE2.backbone_to_site()

        self.cst_idx1 = np.where(self.RE1.atom_names[None, :] == self.csts[:, None])[1]
        self.cst_idx2 = np.where(self.RE2.atom_names[None, :] == self.csts[:, None])[1]
        self.rl1mask = np.argwhere(~np.isin(self.RE1.atom_names, self.csts)).flatten()
        self.rl2mask = np.argwhere(~np.isin(self.RE2.atom_names, self.csts)).flatten()

        self.name = self.res
        if self.site1 is not None:
            self.name = f"{self.RE1.nataa}{self.site1}-{self.RE2.nataa}{self.site2}{self.res}"
        if self.chain is not None:
            self.name += f"_{self.chain}"

        self.selstr = (
            f"resid {self.site1} {self.site2} and segid {self.chain} and not altloc B"
        )

        self._graph = nx.Graph()
        self._graph.add_edges_from(self.bonds)

        self.clash_radius = kwargs.setdefault("clash_radius", None)
        if self.clash_radius is None:
            self.clash_radius = np.linalg.norm(self.clash_ori - self.coords, axis=-1).max() + 5

        self.rmin2 = chilife.get_lj_rmin(self.atom_types[self.side_chain_idx])
        self.eps = chilife.get_lj_eps(self.atom_types[self.side_chain_idx])

        self.protein_setup()
        self.sub_labels = (self.RE1, self.RE2)

    def protein_setup(self):
        if isinstance(self.protein, (mda.AtomGroup, mda.Universe)):
            if not hasattr(self.protein.universe._topology, "altLocs"):
                self.protein.universe.add_TopologyAttr('altLocs', np.full(len(self.protein.universe.atoms), ""))

        self.protein = self.protein.select_atoms("not (byres name OH2 or resname HOH)")
        self.clash_ignore_idx = self.protein.select_atoms(
            f"resid {self.site1} {self.site2} and segid {self.chain}"
        ).ix

        self.resindex = self.protein.select_atoms(self.selstr).residues[0].resindex
        self.segindex = self.protein.select_atoms(self.selstr).residues[0].segindex

        if self.protein_tree is None:
            self.protein_tree = cKDTree(self.protein.atoms.positions)

        protein_clash_idx = self.protein_tree.query_ball_point(
            self.clash_ori, self.clash_radius
        )
        self.protein_clash_idx = [
            idx for idx in protein_clash_idx if idx not in self.clash_ignore_idx
        ]

        _, self.irmin_ij, self.ieps_ij, _ = chilife.prep_internal_clash(self)
        _, self.ermin_ij, self.eeps_ij = chilife.prep_external_clash(self)

        self.aidx, self.bidx = [list(x) for x in zip(*self.non_bonded)]

        if self._minimize:
            self.minimize()

        if self.eval_clash:
            self.evaluate()

    def guess_chain(self):
        if self.protein is None:
            chain = "A"
        elif len(set(self.protein.segments.segids)) == 1:
            chain = self.protein.segments.segids[0]
        elif np.isin(self.protein.residues.resnums, self.site1).sum() == 0:
            raise ValueError(
                f"Residue {self.site1} is not present on the provided protein"
            )
        elif np.isin(self.protein.residues.resnums, self.site1).sum() == 1:
            chain = self.protein.select_atoms(f"resid {self.site1}").segids[0]
        else:
            raise ValueError(
                f"Residue {self.site1} is present on more than one chain. Please specify the desired chain"
            )
        return chain

    def get_lib(self, rotlib):

        # If given a dictionary use that as the rotlib
        if isinstance(rotlib, dict):
            if tuple(sorted(rotlib.keys())) != ('csts', 'libA', 'libB'):
                raise RuntimeError('Non-file dRotamerEnsemble rotlibs must be dictionaries consisting of exactle 3 '
                                   'entries: `csts`, `libA` and `libB`')
            self.csts = rotlib['csts']
            self.libA, self.libB = rotlib['libA'], rotlib['libB']
            self.kwargs["eval_clash"] = False
            return None

        rotlib = self.res if rotlib is None else rotlib
        if 'ip' not in rotlib:
            rotlib += f'ip{self.increment}'

        # Check if any exist
        rotlib_path = chilife.get_possible_rotlibs(rotlib, suffix='drotlib', extension='.zip')

        if rotlib_path is None:
            # Check if libraries exist but for different i+n
            rotlib_path = chilife.get_possible_rotlibs(rotlib.replace(f'ip{self.increment}', ''),
                                                       suffix='drotlib',
                                                       extension='.zip',
                                                       return_all=True)

            if rotlib_path is None:
                raise NameError(f'There is no rotamer library called {rotlib} in this directory or in chilife')

            else:
                warnings.warn(f'No rotamer library found for the given increment (ip{self.increment}) but rotlibs '
                              f'were found for other increments. chiLife will combine these rotlib to model '
                              f'this site1 pair and but they may not be accurate! Because there is no information '
                              f'about the relative weighting of different rotamer libraries all weights will be '
                              f'set to 1/len(rotlib)')

        if isinstance(rotlib_path, Path):
            libA, libB, csts = chilife.read_library(rotlib_path)

        elif isinstance(rotlib_path, list):
            concatable = ['dihedrals', 'internal_coords']

            cctA, cctB, ccsts = {}, {}, {}
            libA, libB, csts = chilife.read_library(rotlib_path[0])
            unis = []

            for lib in (libA, libB):
                names, types = lib['atom_names'], libA['atom_types']
                residxs = np.zeros(len(names), dtype=int)
                resnames, resids = np.array([self.res]), np.array([1])
                segidx = np.array([0])

                uni = chilife.make_mda_uni(names, types, resnames, residxs, resids, segidx)
                unis.append(uni)

            for p in rotlib_path:
                tlibA, tlibB, tcsts = chilife.read_library(p)
                for lib, tlib, cct, uni in zip((libA, libB), (tlibA, tlibB), (cctA, cctB), unis):

                    # Libraries must have the same atom order
                    if not np.all(np.isin(tlib['atom_names'], lib['atom_names'])) and \
                            np.all(tlib['dihedral_atoms'] == lib['dihedral_atoms']):
                        raise ValueError(f'Rotlibs {rotlib_path[0].stem} and {p.stem} are not compatable. You may'
                                         f'need to rename one of them.')

                    # Map coordinates
                    ixmap = [np.argwhere(tlib['atom_names'] == aname).flat[0] for aname in lib['atom_names']]
                    cct.setdefault('coords', []).append(tlib['coords'][:, ixmap])

                    # Create new internal coords if they are defined differently
                    lib_ic, tlib_ic = lib['internal_coords'][0], tlib['internal_coords'][0]
                    if np.any(lib_ic.atom_names != tlib_ic.atom_names):
                        uni.load_new(cct['coords'][-1])
                        ics = [chilife.get_internal_coords(uni, lib['dihedral_atoms'], lib_ic.bonded_pairs)
                               for ts in uni.trajectory]
                        tlib['internal_coords'] = ics

                    for field in concatable:
                        cct.setdefault(field, []).append(tlib[field])

                    for field in concatable + ['coords']:
                        lib[field] = np.concatenate(cct[field])

            libA['weights'] = libB['weights'] = np.ones(len(libA['coords'])) / len(libA['coords'])

        self.csts = csts
        self.libA, self.libB = libA, libB
        self.kwargs["eval_clash"] = False

    def create_ensembles(self):

        self.RE1 = chilife.RotamerEnsemble(self.res,
                                           self.site1,
                                           self.protein,
                                           self.chain,
                                           self.libA,
                                           **self.kwargs)

        self.RE2 = chilife.RotamerEnsemble(self.res,
                                           self.site2,
                                           self.protein,
                                           self.chain,
                                           self.libB,
                                           **self.kwargs)

    def save_pdb(self, name=None):
        if name is None:
            name = self.name + ".pdb"
        if not name.endswith(".pdb"):
            name += ".pdb"

        chilife.save(name, self.RE1, self.RE2)

    def minimize(self, callback=None):

        scores = [self._min_one(i, ic1, ic2, callback=callback) for i, (ic1, ic2) in
                  enumerate(zip(self.RE1.internal_coords, self.RE2.internal_coords))]

        scores = np.asarray(scores)
        SSEs = np.linalg.norm(self.RE1.coords[:, self.cst_idx1] - self.RE2.coords[:, self.cst_idx2], axis=2).sum(axis=1)
        MSD = SSEs / len(self.csts)
        MSDmin = MSD.min()

        if MSDmin > 0.1:
            warnings.warn(f'The minimum MSD of the cap is {MSD.min()}, this may result in distorted spin label. '
                          f'Check that the structures make sense.')

        if MSDmin > 0.25:
            raise RuntimeError(f'chiLife was unable to connect residues {self.site1} and {self.site2} with {self.res}. '
                               f'Please double check that this is the intended labeling site. It is likely that these '
                               f'sites are too far apart.')
        self.cap_MSDs = MSD
        self.RE1.backbone_to_site()
        self.RE2.backbone_to_site()

        scores -= scores.min()
        self.rotamer_scores = scores
        self.weights *= np.exp(-scores / (chilife.GAS_CONST * self.temp) / np.exp(-scores).sum())
        self.weights /= self.weights.sum()

    def _objective(self, dihedrals, ic1, ic2):

        ic1.set_dihedral(dihedrals[: len(self.RE1.dihedral_atoms)], 1, self.RE1.dihedral_atoms)
        coords1 = ic1.to_cartesian()[self.RE1.ic_mask]

        ic2.set_dihedral(dihedrals[-len(self.RE2.dihedral_atoms):], 1, self.RE2.dihedral_atoms)
        coords2 = ic2.to_cartesian()[self.RE2.ic_mask]

        diff = np.linalg.norm(coords1[self.cst_idx1] - coords2[self.cst_idx2], axis=1)
        ovlp = (coords1[self.cst_idx1] + coords2[self.cst_idx2]) / 2
        coords = np.concatenate([coords1[self.rl1mask], coords2[self.rl2mask], ovlp], axis=0)
        r = np.linalg.norm(coords[self.aidx] - coords[self.bidx], axis=1)

        # Faster to compute lj here
        lj = self.irmin_ij / r
        lj = lj * lj * lj
        lj = lj * lj

        # attractive forces are needed, otherwise this term will perpetually push atoms apart
        internal_energy = self.ieps_ij * (lj * lj - 2 * lj)
        score = (diff @ diff) * self.restraint_weight / len(diff) + internal_energy.sum()

        return score

    def _min_one(self, i, ic1, ic2, callback=None):
        if callback is not None:
            if 'i' in inspect.signature(callback).parameters:
                callback = partial(callback, i=i)


        d0 = np.concatenate([ic1.get_dihedral(1, self.RE1.dihedral_atoms),
                             ic2.get_dihedral(1, self.RE2.dihedral_atoms)])

        lb = d0 - np.pi  # np.deg2rad(40)
        ub = d0 + np.pi  # np.deg2rad(40) #
        bounds = np.c_[lb, ub]
        xopt = opt.minimize(self._objective, x0=d0, args=(ic1, ic2),
                            bounds=bounds, method=self.min_method,
                            callback=callback)
        self.RE1._coords[i] = ic1.coords[self.RE1.H_mask]
        self.RE2._coords[i] = ic2.coords[self.RE2.H_mask]
        tors = d0 - xopt.x
        tors = np.arctan2(np.sin(tors), np.cos(tors))
        tors = np.sqrt(tors @ tors)

        return xopt.fun + tors * self.torsion_weight

    @property
    def weights(self):
        return self.RE1.weights

    @weights.setter
    def weights(self, value):
        self.RE1.weights = value
        self.RE2.weights = value

    @property
    def coords(self):
        ovlp = (self.RE1.coords[:, self.cst_idx1] + self.RE2.coords[:, self.cst_idx2]) / 2
        return np.concatenate([self.RE1._coords[:, self.rl1mask], self.RE2._coords[:, self.rl2mask], ovlp], axis=1)

    @coords.setter
    def coords(self, value):
        if value.shape[1] != len(self.atom_names):
            raise ValueError(
                f"The provided coordinates do not match the number of atoms of this ensemble ({self.res})"
            )

        self.RE1._coords[:, self.rl1mask] = value[:, :len(self.rl1mask)]
        self.RE2._coords[:, self.rl2mask] = value[:, len(self.rl1mask):len(self.rl1mask) + len(self.rl2mask)]
        self.RE1._coords[:, self.cst_idx1] = value[:, len(self.rl1mask) + len(self.rl2mask):]
        self.RE2._coords[:, self.cst_idx2] = value[:, len(self.rl1mask) + len(self.rl2mask):]

    @property
    def _lib_coords(self):
        ovlp = (self.RE1._lib_coords[:, self.cst_idx1] + self.RE2._lib_coords[:, self.cst_idx2]) / 2
        return np.concatenate([self.RE1._lib_coords[:, self.rl1mask],
                               self.RE2._lib_coords[:, self.rl2mask], ovlp], axis=1)

    @property
    def atom_names(self):
        return np.concatenate((self.RE1.atom_names[self.rl1mask],
                               self.RE2.atom_names[self.rl2mask],
                               self.RE1.atom_names[self.cst_idx1]))

    @property
    def atom_types(self):
        return np.concatenate((self.RE1.atom_types[self.rl1mask],
                               self.RE2.atom_types[self.rl2mask],
                               self.RE1.atom_types[self.cst_idx2]))

    @property
    def centroid(self):
        return self.coords.mean(axis=(0, 1))

    @property
    def clash_ori(self):

        if isinstance(self._clash_ori_inp, (np.ndarray, list)):
            if len(self._clash_ori_inp) == 3:
                return self._clash_ori_inp

        elif isinstance(self._clash_ori_inp, str):
            if self._clash_ori_inp in ["cen", "centroid"]:
                return self.centroid

            elif (ori_name := self._clash_ori_inp.upper()) in self.atom_names:
                return np.squeeze(self.coords[0][ori_name == self.atom_names])

        else:
            raise ValueError(
                f"Unrecognized clash_ori option {self._clash_ori_inp}. Please specify a 3D vector, an "
                f"atom name or `centroid`"
            )

        return self._clash_ori

    @clash_ori.setter
    def clash_ori(self, inp):
        self._clash_ori_inp = inp

    @property
    def side_chain_idx(self):
        "Indices of side chain atoms"
        if not hasattr(self, '_side_chain_idx'):
            self._side_chain_idx = np.argwhere(
                np.isin(self.atom_names, dRotamerEnsemble.backbone_atoms, invert=True)
            ).flatten()

        return self._side_chain_idx

    @property
    def bonds(self):
        """List of intra-label atom pairs that are covalently bonded."""
        if not hasattr(self, "_bonds"):
            bonds = []

            for bond in self.RE1.bonds:
                bndin = np.isin(bond, self.rl1mask)
                if np.all(bndin):
                    bonds.append(bond)
                elif np.any(bndin):
                    bonds.append([bond[0], np.argwhere(self.atom_names == self.RE1.atom_names[bond[1]]).flat[0]])
                else:
                    bonds.append([np.argwhere(self.atom_names == self.RE1.atom_names[bond[0]]).flat[0],
                                  np.argwhere(self.atom_names == self.RE1.atom_names[bond[1]]).flat[0]])

            for bond in self.RE2.bonds:
                bndin = np.isin(bond, self.rl2mask)
                if np.all(bndin):
                    bonds.append([b + len(self.rl1mask) for b in bond])
                elif not bndin[1]:
                    bonds.append([bond[0] + len(self.rl1mask),
                                  np.argwhere(self.atom_names == self.RE2.atom_names[bond[1]]).flat[0]])

            self._bonds = np.array(sorted(set(map(tuple, bonds))), dtype=int)

        return self._bonds

    @bonds.setter
    def bonds(self, inp):
        """
        Set of intralabel bonded pairs.
        Parameters
        ----------
        inp : ArrayLike
            List of atom ID pairs that are bonded
        """
        self._bonds = set(tuple(i) for i in inp)
        idxs = np.arange(len(self.atom_names))
        all_pairs = set(combinations(idxs, 2))
        self._non_bonded = all_pairs - self._bonds

    @property
    def non_bonded(self):
        """ list of indices of intra-label non-bonded atom pairs. Primarily used for internal clash evaluation when
        sampling the dihedral space"""
        if not hasattr(self, "_non_bonded"):
            pairs = dict(nx.all_pairs_shortest_path(self._graph, self._exclude_nb_interactions - 1))
            pairs = {(a, b) for a in pairs for b in pairs[a] if a < b}
            all_pairs = set(combinations(range(len(self.atom_names)), 2))
            self._non_bonded = all_pairs - pairs

        return sorted(list(self._non_bonded))

    @non_bonded.setter
    def non_bonded(self, inp):
        """
        Create a set of non-bonded atom pair indices within a single side chain
        Parameters
        ----------
        inp : ArrayLike
            List of atom ID pairs that are not bonded
        """
        self._non_bonded = set(tuple(i) for i in inp)
        idxs = np.arange(len(self.atom_names))
        all_pairs = set(combinations(idxs, 2))
        self._bonds = all_pairs - self._non_bonded

    def trim_rotamers(self):
        """Remove low probability rotamers from the ensemble"""

        arg_sort_weights = np.argsort(self.weights)[::-1]
        sorted_weights = self.weights[arg_sort_weights]
        cumulative_weights = np.cumsum(sorted_weights)
        cutoff = np.maximum(1, len(cumulative_weights[cumulative_weights < 1 - self.trim_tol]))
        keep_idx = arg_sort_weights[:cutoff]
        if hasattr(self, 'cap_MSDs'):
            self.cap_MSDs = self.cap_MSDs[keep_idx]
            self.rotamer_scores = self.rotamer_scores[keep_idx]

        if hasattr(self, 'rot_clash_energy'):
            self.rot_clash_energy = self.rot_clash_energy[keep_idx]
        self.RE1.trim_rotamers(keep_idx=keep_idx)
        self.RE2.trim_rotamers(keep_idx=keep_idx)

    def evaluate(self):
        """Place rotamer ensemble on protein site1 and recalculate rotamer weights."""
        # Calculate external energies
        energies = self.energy_func(self)
        self.rot_clash_energy = energies
        # Calculate total weights (combining internal and external)
        self.weights, self.partition = chilife.reweight_rotamers(energies, self.temp, self.weights)
        logging.info(f"Relative partition function: {self.partition:.3}")

        # Remove low-weight rotamers from ensemble
        if self._do_trim:
            self.trim_rotamers()

    def __len__(self):
        """Return the length of the rotamer library (number of rotamers)"""
        return len(self.RE1.coords)

    def copy(self):
        """Create a deep copy of the dRotamerEnsemble object"""
        new_copy = chilife.dRotamerEnsemble(self.res, (self.site1, self.site2), chain=self.chain,
                                            protein=self.protein,
                                            rotlib={'csts': self.csts, 'libA': self.libA, 'libB': self.libB},
                                            minimize=False,
                                            eval_clash=False)
        for item in self.__dict__:
            if isinstance(self.dict[item], np.ndarray):
                new_copy.__dict__[item] = self.__dict__[item].copy()

            elif item == 'RE1':
                new_copy.__dict__[item] == self.__dict__[item].copy(rotlib=self.libA)

            elif item == 'RE2':
                new_copy.__dict__[item] == self.__dict__[item].copy(rotlib=self.libB)

            elif item == 'protein':
                pass

            else:
                new_copy.__dict__[item] = deepcopy(self.__dict__[item])
        return new_copy



from functools import partial
import numpy as np
import pandas as pd
import pickle
from glob import glob
import matplotlib.pyplot as plt
import os
from tqdm import tqdm
import multiprocessing as mp
from mpl_toolkits.mplot3d import Axes3D


"""This utility will construct all rotamer library in the proper coordinate frame when provided with 1) and deer2bb 
rotamer library in pickle format (Same as pymol) and 2) pdb files with the base structure of all rotamers of interest
PDB files and deer2bb_bb_ind.pkl names must be the same with the exception of the pdb extension. This application also
will plot the NO midpoint coordinates and N, CA, CB coordinates so that the user can review the coordinate 
transformations to ensure nothing went wrong"""


def main():
    if os.path.exists('file.pdb'):
        os.remove('file.pdb')
    names = ['resn', 'Phi', 'Psi', 'count', 'r1', 'r2', 'r3', 'r4', 'r5', 'Prob',
             'chi1Val', 'chi2Val', 'chi3Val', 'chi4Val', 'chi5Val', 'chi1Sig',
             'chi2Sig', 'chi3Sig', 'chi4Sig', 'chi5Sig']
    bbdep_lib = pd.read_csv('R1C.rotlib', delim_whitespace=True, names=names)
    bbdep_lib['resn'] = 'R1C'
    bbdep_lib_dict = {}
    rotamer_name = 'R1C'
    pdb_filename = rotamer_name + '.pdb'
    col_names = ['ATOM', 'atom number', 'atom name', 'residue name', 'chain', 'residue number',
                 'x', 'y', 'z', 'occupancy', 'Bfactor', 'atom type']
    coords = ['x', 'y', 'z']
    print(pdb_filename)
    struct = pd.read_csv(pdb_filename,
                         dtype={'ATOM': str, 'atom number': int, 'atom name': str, 'residue name': str,
                                'chain': str, 'residue number': int, 'x': float, 'y': float, 'z': float,
                                'occupancy': float, 'Bfactor': float, 'atom type': str},
                         delim_whitespace=True, names=col_names)

    atom_names = struct['atom name'].values
    atom_type = struct['atom type'].values
    dihedral_definitions = [('N', 'CA', 'CB', 'SG'),
                            ('CA', 'CB', 'SG', 'SD'),
                            ('CB', 'SG', 'SD', 'CE'),
                            ('SG', 'SD', 'CE', 'C3'),
                            ('SD', 'CE', 'C3', 'C2')]

    # Align rotamer backbone to defined coordinate system
    N = struct[struct['atom name'] == 'N'][coords].values[0]
    C = struct[struct['atom name'] == 'C'][coords].values[0]
    CA = struct[struct['atom name'] == 'CA'][coords].values[0]

    ori, mx = rosetta_local_mx(N, CA, C)

    struct[coords] -= ori
    struct[coords] = struct[coords].values.dot(mx)

    groups = bbdep_lib.groupby(['resn', 'Phi', 'Psi'])
    with mp.Pool(processes=mp.cpu_count() - 2) as pool:
        libs = tqdm(pool.imap(partial(all_structs,
                          struct=struct,
                          dihedral_definitions=dihedral_definitions,
                          coords=coords), groups), total=1369)

        lib = {results[0]: results[1:] for results in libs}

    np.savez(rotamer_name + '_bbdep_rotlib.npz', lib=lib, atom_types=atom_type,
              atom_names=atom_names, dihedral_atoms=dihedral_definitions)

def all_structs(grp, struct, dihedral_definitions, coords):
    key, grp = grp
    weights = grp.Prob.values
    rot_lib = grp[['chi1Val', 'chi2Val', 'chi3Val', 'chi4Val', 'chi5Val']].values
    rotamer_coords = [get_struct(rotamer, struct, dihedral_definitions, coords) for rotamer in rot_lib]
    rotamer_coords = np.array(rotamer_coords)
    return key, rot_lib, rotamer_coords, weights

def get_struct(rotamer, struct0, dihedral_definitions, coords):
    struct = struct0.copy()
    for angle, dihedral in zip(rotamer, dihedral_definitions):
        mobile_idx = int(struct.index.where(struct['atom name'] == dihedral[-2]).dropna().values[0])
        mobile = struct.iloc[mobile_idx:-2, 6:9].values
        p = struct[struct['atom name'].isin(dihedral)][coords].values
        struct.iloc[mobile_idx:-2, 6:9] = set_dihedral(p, angle, mobile)

    return struct[coords].values

def rosetta_local_mx(N, CA, C):
    """
    Calculates a translation and rotation matrix to transform an amino acid from the global coordinate frame to the
    local coordinate frame as defined by the rosetta convention. The analogous transformation can be found in:
    src/numeric/HomogeneousTransform.cc

    :param N: numpy ndarray (1x3)
        Backbone nitrogen coordinates

    :param CA: numpy ndarray (1x3)
        Backbone Calpha carbon coordinates

    :param C: numpy ndarray (1x3)
        Backbone carbonyl carbon coordinates

    :return (origin, rotation_matrix) : (numpy ndarray (1x3), numpy ndarra (3x3))
        origin: new origin position in 3 dimensional space
        rotation_matrix: rotation  matrix to rotate spin label to
    """

    # Tranform coordinates such that the CA atom is at the origin
    p1 = N - CA
    p2 = C - CA
    p3 = CA - CA

    # Define Z-axis unit vector
    zaxis = p3 - p2
    zaxis = zaxis / np.linalg.norm(zaxis)

    # Calculate C-->N vector
    p21 = p1 - p2

    # Calculate magnitude of C-->N vector in Z direction
    p21_z_comp = p21.dot(zaxis)

    # Subtract Z component from C-->N vector to get a vector orthogonal to Z direction in the N-Ca-C plane
    yaxis = p21 - p21_z_comp * zaxis
    yaxis = yaxis / np.linalg.norm(yaxis)

    # Calculate x axis as cross product of y and z axis
    xaxis = np.cross(yaxis, zaxis)

    # Create rotation matrix to rotate any atoms to new coordinate frame
    rotation_matrix = np.linalg.inv(np.array([xaxis, yaxis, zaxis]))

    # Set origin at calpha
    origin = CA

    return origin, rotation_matrix


def get_dihedral(p):
    """
    Calculates dihedral of a given set of atoms, p = [0, 2, 3, 3] using Praxeolitic's formula (stackoverflow). Returns
    value in degrees.

                    3
         ------>  /
        1-------2
      /
    0
    """

    p0 = p[0]
    p1 = p[1]
    p2 = p[2]
    p3 = p[3]

    # Define vectors from coordinates
    b0 = -1.0 * (p1 - p0)
    b1 = p2 - p1
    b2 = p3 - p2

    # Normalize dihedral bond vector
    b1 /= np.linalg.norm(b1)

    # Calculate dihedral projections orthogonal to the bond vector
    v = b0 - np.dot(b0, b1) * b1
    w = b2 - np.dot(b2, b1) * b1

    # Calculate angle between projections
    x = np.dot(v, w)
    y = np.dot(np.cross(b1, v), w)

    # Return as degrees
    return np.degrees(np.arctan2(y, x))


def set_dihedral(p, angle, mobile):
    """
    Sets the dihedral angle by rotating all 'mobile' atoms from their current position about the dihedral bond defined
    by the four atoms in p. Dihedral will be set to the value of 'angle' in degrees

    :param p: array-like int
        Atom coordinates that define dihedral to rotate about

    :param angle: float
        New angle to set the dihedral to (degrees).

    :param mobile
        atoms to move by setting dihedral

    """
    current = get_dihedral(p)
    angle = angle - current
    angle = np.deg2rad(angle)

    ori = p[1]
    mobile -= ori
    v = p[2] - p[1]
    v /= np.linalg.norm(v)
    R = get_dihedral_rotation_matrix(angle, v)

    new_mobile = R.dot(mobile.T).T + ori

    return new_mobile


def get_dihedral_rotation_matrix(theta, v):
    """
    Build a matrix that will rotate any coordinates about a vector, v, by theta in radians

    :param theta: float
        Rotation  angle in radians

    :param v: numpy ndarray (1x3)
        Three dimensional vector to rotate about

    :return rotation_matrix:
        Matrix that will rotate
    """

    # Normalize input vector
    v = v / np.linalg.norm(v)

    # Compute Vx matrix
    Vx = np.zeros((3, 3))
    Vx[[2, 0, 1], [1, 2, 0]] = v
    Vx -= Vx.T

    # Rotation matrix. See https://en.wikipedia.org/wiki/Rotation_matrix#Rotation_matrix_from_axis_and_angle
    rotation_matrix = np.identity(3) * np.cos(theta) + np.sin(theta) * Vx + (1 - np.cos(theta)) * np.outer(v, v)

    return rotation_matrix

pdb_formatter = {'ATOM': '{:<5s}'.format, 'atom number': '{:>5d}'.format, 'atom name': '{:<4s} '.format,
                 'residue name': '{:<3s}'.format, 'chain': '{:<1s}'.format, 'residue number': '{:>4d} '.format,
                 'x': '{:>8.3f}'.format, 'y': '{:>8.3f}'.format, 'z': '{:>8.3f}'.format, 'occupancy': '{:>6.2f}'.format,
                 'Bfactor': '{:>6.2f}'.format, 'atom type': '{:>2s}'.format}

if __name__ == '__main__':
    main()



import hashlib, os, pickle
from functools import partial

import matplotlib.pyplot as plt
import numpy as np
import MDAnalysis as mda
import pytest
import chilife

labels = ["R1M", "R7M", "V1M", "M1M", "I1M"]
ubq = mda.Universe("test_data/1ubq.pdb", in_memory=True)
U = mda.Universe("test_data/1omp.pdb", in_memory=True)
protein = U.select_atoms("protein")

hashes = {}
with open("test_data/hashes.txt", "r") as f:
    for line in f:
        if len(line.split()) == 2:
            a, b = line.split()
            hashes[a] = b
efunc = partial(chilife.get_lj_rep, forgive=0.8)
kwinput = (
    [{"label": l, "energy_func": efunc} for l in labels]
    + [{"label": l, "site": 25, "chain": "A", "energy_func": efunc} for l in labels]
    + [
        {"label": l, "site": 25, "chain": "A", "protein": protein, "energy_func": efunc}
        for l in labels
    ]
    + [
        {"label": l, "site": 25, "chain": "A", "protein": U, "energy_func": efunc}
        for l in labels
    ]
)


def test_bbdep_construction1():
    efunc = partial(chilife.get_lj_energy, cap=np.inf)
    SL1 = chilife.SpinLabel(
        "R1C", site=211, chain="A", protein=U, energy_func=efunc, forgive=0.5
    )
    SL2 = chilife.SpinLabel(
        "R1C", site=295, chain="A", protein=U, energy_func=efunc, forgive=0.5
    )

    chilife.save("bbdep.pdb", SL1, SL2, protein_path="test_data/1omp.pdb")

    with open("bbdep.pdb", "r") as f:
        test = hashlib.md5(f.read().encode('utf-8')).hexdigest()

    with open("test_data/bbdep.pdb", "r") as f:
        ans = hashlib.md5(f.read().encode('utf-8')).hexdigest()

    os.remove("bbdep.pdb")

    assert test == ans


@pytest.mark.parametrize("label", labels)
def test_eq(label):
    SL1 = chilife.SpinLabel(label)
    SL2 = chilife.SpinLabel(label, site=25)
    SL3 = chilife.SpinLabel(label, site=25, chain="A")
    SL4 = chilife.SpinLabel(label, site=25, chain="A", protein=protein)
    assert SL1 == SL2 == SL3 != SL4


@pytest.mark.parametrize("label", labels)
def test_lib2site(label):
    site = np.array([[-7.882, -3.264, -26.954],
                     [-8.528, -3.130, -28.260],
                     [-7.611, -2.318, -29.157]])

    lib = chilife.SpinLabel(label)
    lib.to_site(site)
    ans = np.load(f"test_data/{label}_lib2site.npy")
    np.testing.assert_almost_equal(ans, lib._coords)


@pytest.mark.parametrize("label", labels)
def test_add_protein(label):

    site1, chain = 25, "A"
    lib = chilife.SpinLabel(label, site1, chain=chain)
    lib.protein = U
    lib.protein_setup()

    with np.load(f"test_data/{label}_label.npz") as f:
        np.testing.assert_almost_equal(f["coords"], lib._coords)
        np.testing.assert_almost_equal(f["weights"], lib.weights)


def test_from_wizard():
    np.random.seed(200)
    SL = chilife.SpinLabel.from_wizard("TRT", 28, ubq, to_find=10)
    for ic, dihedral in zip(SL.internal_coords, SL.dihedrals):
        np.testing.assert_allclose(np.rad2deg(ic.get_dihedral(1, SL.dihedral_atoms)), dihedral)

    with open("test_data/from_wiz.pkl", "rb") as f:
        SLans = pickle.load(f)

    np.testing.assert_almost_equal(SL.coords, SLans.coords, decimal=5)
    np.testing.assert_almost_equal(SL.weights, SLans.weights)
    np.testing.assert_almost_equal(SL.dihedrals, SLans.dihedrals, decimal=5)
    assert len(SL) == len(SL.dihedrals) == len(SL.weights) == len(SL.coords)


def test_minimize():

    SL1 = chilife.SpinLabel("R1M", 20, protein=U, minimize=True)
    SL2 = chilife.SpinLabel("R1M", 238, protein=U, minimize=True)

    r = np.linspace(15, 80, 256)
    P = chilife.distance_distribution(SL1, SL2, r)
    assert np.max(P) - 0.20643375571027256 <= 1e-2
    assert np.argmax(P) == 65


def test_spin_center_array_dim():
    SL1 = chilife.SpinLabel('R1M', 5, ubq)
    assert SL1.spin_centers.shape == (1, 3)

def test_no_altoloc():
    GR1G = mda.Universe('test_data/GR1G.gro')
    assert not hasattr(GR1G._topology, "altLocs")
    SL = chilife.SpinLabel('R1M', 2, GR1G)
    assert hasattr(GR1G._topology, "altLocs")





import MDAnalysis as mda
from pathlib import Path
import chilife
import pickle

pdb_directory = Path('../residue_pdbs')
residue_pdbs = pdb_directory.glob('*.pdb')

for pdb in residue_pdbs:
    print(pdb.stem)

    srtd = chilife.sort_pdb(pdb)
    with open(pdb, 'w') as f:
        f.writelines(srtd)

    struct = mda.Universe(str(pdb))
    ICs = chilife.get_internal_coords(struct)
    with open(pdb.stem + '_ic.pkl', 'wb') as f:
        pickle.dump(ICs, f)


import pickle, json
import numpy as np
import chilife as xl
from pathlib import Path
import zipfile, os

with open('../spin_atoms.json', 'r') as f:
    spin_atoms = json.load(f)


def convert_library(libname):
    with open(f'../residue_internal_coords/{libname}C_ic.pkl', 'rb') as f:
        cst_idxs, csts = pickle.load(f)

    save_dict = {'cst_pairs': cst_idxs, 'cst_distances': csts}
    np.savez(f'{libname}_csts.npz', **save_dict)

    with zipfile.ZipFile(f'{libname}_drotlib.zip', mode='w') as archive:
        archive.write(f'{libname}A_rotlib.npz')
        archive.write(f'{libname}B_rotlib.npz')
        archive.write(f'{libname}_csts.npz')

        # Cleanup intermediate files
    # os.remove(f'{libname}A_rotlib.npz')
    # os.remove(f'{libname}B_rotlib.npz')
    # os.remove(f'{libname}_csts.npz')


    # np.savez(library, **lib)

def add_to_library(libname):

    with zipfile.ZipFile(libname + '_drotlib.zip', 'r') as archive:
        archive.extractall()
    files = list(Path.cwd().glob(f'{libname}*.npz'))
    for file in files:
        if 'csts' not in file.name:
            with np.load(file, allow_pickle=True) as f:
                lib = dict(f)

            resname = libname[:3]

            lib['rotlib'] = libname + file.name[len(libname)]
            print(lib['rotlib'])
            lib['resname'] = resname
            spin_atoms = {'spin_atoms': ['Cu1'], 'spin_weights': ['1.0']}

            lib.update(spin_atoms)

            lib['type'] = 'chilife rotamer library'
            lib['format_version'] = 1.0

            np.savez(file, **lib)

    with zipfile.ZipFile(f'{libname}_drotlib.zip', mode='w') as archive:
        archive.write(f'{libname}A_rotlib.npz')
        archive.write(f'{libname}B_rotlib.npz')
        archive.write(f'{libname}_csts.npz')

    # Cleanup intermediate files
    os.remove(f'{libname}A_rotlib.npz')
    os.remove(f'{libname}B_rotlib.npz')
    os.remove(f'{libname}_csts.npz')


def adjust_library(libname):

    with zipfile.ZipFile(libname + '_drotlib.zip', 'r') as archive:
        archive.extractall()

    files = list(Path.cwd().glob(f'{libname}*.npz'))
    for file in files:
        if 'csts' not in file.name:
            with np.load(file, allow_pickle=True) as f:
                lib = dict(f)

                lib['spin_weights'] = lib['spin_weights'].astype(float)

            np.savez(file, **lib)

    with zipfile.ZipFile(f'{libname}_drotlib.zip', mode='w') as archive:
        archive.write(f'{libname}A_rotlib.npz')
        archive.write(f'{libname}B_rotlib.npz')
        archive.write(f'{libname}_csts.npz')

    # Cleanup intermediate files
    os.remove(f'{libname}A_rotlib.npz')
    os.remove(f'{libname}B_rotlib.npz')
    os.remove(f'{libname}_csts.npz')



# convert_library('DHCip2')
# convert_library('DHCip4')

# add_to_library('DHCip2')
# add_to_library('DHCip4')


adjust_library('DHCip2')
adjust_library('DHCip4')


import os, hashlib
import time
from functools import partial
import numpy as np
import pytest
import MDAnalysis as mda
import chilife

pdbids = ["1ubq", "1a2w", '1az5']
ubq = mda.Universe("test_data/1ubq.pdb", in_memory=True)
resis = [
    (key, -90, 160)
    for key in chilife.dihedral_defs
    if key
    not in (("R1B", "R1C", "CYR1", "MTN", "ALA", "GLY") +
            tuple(chilife.USER_LIBRARIES) +
            tuple(chilife.USER_dLIBRARIES))
]


@pytest.mark.parametrize("res", resis)
def test_read_dunbrack(res):
    res, phi, psi = res

    dlib = chilife.read_bbdep(res, -70, 90)

    dlib_mx, dlib_ori = chilife.global_mx(*np.squeeze(dlib["coords"][0, :3]))
    dlib["coords"] = np.einsum("ijk,kl->ijl", dlib["coords"], dlib_mx) + dlib_ori

    with np.load(f"test_data/{res}_{phi}_{psi}.npz", allow_pickle=True) as f:
        dlib_ans = {key: f[key] for key in f if key != "allow_pickle"}

    for key in dlib_ans:
        if dlib_ans[key].dtype not in [np.dtype(f"<U{i}") for i in range(1, 5)]:
            np.testing.assert_almost_equal(dlib_ans[key], dlib[key])
        else:
            assert np.all(np.char.equal(dlib_ans[key], dlib[key]))


def test_sort_pdb():
    pdbfile = "test_data/trt.pdb"
    x = chilife.sort_pdb(pdbfile)

    linest = [line for line in x if line[76:79].strip() != 'H']
    test = hashlib.md5("".join(linest).encode("utf-8")).hexdigest()

    with open("test_data/trt_sorted.pdb", "r") as f:
        lines = f.readlines()
        linesa = [line for line in lines if line[76:79].strip() != 'H']
        ans = hashlib.md5("".join(linesa).encode("utf-8")).hexdigest()

    assert test == ans


def test_sort_pdb2():
    x = chilife.sort_pdb("test_data/SL_GGAGG.pdb")

    with open("test_data/SL_GGAGG_tmp.pdb", "w") as f:
        for line in x:
            f.write(line)

    U = mda.Universe("test_data/SL_GGAGG_tmp.pdb", in_memory=True)
    os.remove("test_data/SL_GGAGG_tmp.pdb")
    ICs = chilife.get_internal_coords(U, preferred_dihedrals=[["C", "N", "CA", "C"]])


def test_sort_H():
    x = chilife.sort_pdb('../chilife/data/rotamer_libraries/residue_pdbs/lys.pdb')
    H_names = [xx[12:16] for xx in x[9:]]
    np.testing.assert_equal(H_names, [' H  ', ' HA ', '3HB ', '2HB ',
                                      '3HG ', '2HG ', '3HD ', '2HD ',
                                      '2HE ', '3HE ', '3HZ ', '2HZ ',
                                      '1HZ '])

def test_sort_manymodels():
    x = chilife.sort_pdb("test_data/msort.pdb")
    with open("test_data/msort_tmp.pdb", "w") as f:
        for i, struct in enumerate(x):
            f.write(f"MODEL {i}\n")
            f.writelines(struct)
            f.write("ENDMDL\n")

    with open("test_data/msort_tmp.pdb", "r") as f:
        heavy_test = [line for line in f.readlines() if line[76:79].strip() != 'H']

        test = hashlib.md5("".join(heavy_test).encode("utf-8")).hexdigest()

    with open("test_data/msort_ans.pdb", "r") as f:
        heavy_ans = [line for line in f.readlines() if line[76:79].strip() != 'H']

        ans = hashlib.md5("".join(heavy_ans).encode("utf-8")).hexdigest()

    os.remove('test_data/msort_tmp.pdb')

    assert test == ans


def test_mutate():
    protein = mda.Universe("test_data/1ubq.pdb", in_memory=True).select_atoms("protein")
    SL = chilife.SpinLabel(
        "R1C",
        site=28,
        protein=protein,
        energy_func=partial(chilife.get_lj_rep, forgive=0.8),
    )

    labeled_protein = chilife.mutate(protein, SL)
    ub1_A28R1 = mda.Universe("test_data/1ubq_A28R1.pdb", in_memory=True)

    np.testing.assert_almost_equal(ub1_A28R1.atoms.positions, labeled_protein.atoms.positions, decimal=3)


def test_mutate2():
    protein = mda.Universe('test_data/1ubq.pdb', in_memory=True).select_atoms("protein")
    SL1 = chilife.SpinLabel(
        "R1C",
        site=28,
        protein=protein,
        energy_func=partial(chilife.get_lj_rep, forgive=0.8),
    )
    SL2 = chilife.SpinLabel(
        "R1C",
        site=48,
        protein=protein,
        energy_func=partial(chilife.get_lj_rep, forgive=0.8),
    )

    labeled_protein = chilife.mutate(protein, SL1, SL2)
    ub1_A28R1_K48R1 = mda.Universe("test_data/ub1_A28R1_K48R1.pdb")
    np.testing.assert_almost_equal(ub1_A28R1_K48R1.atoms.positions, labeled_protein.atoms.positions, decimal=3)


def test_mutate3():
    protein = mda.Universe("test_data/1omp_H.pdb").select_atoms("protein")
    SL1 = chilife.SpinLabel("R1C", site=238, protein=protein, use_H=True)
    SL2 = chilife.SpinLabel("R1C", site=311, protein=protein, use_H=True)

    labeled_protein = chilife.mutate(protein, SL1, SL2)
    assert len(labeled_protein.atoms) != len(protein.atoms)


def test_mutate4():
    protein = mda.Universe("test_data/1omp_H.pdb").select_atoms("protein")
    D41G = chilife.RotamerEnsemble("GLY", 41, protein=protein)
    S238A = chilife.RotamerEnsemble("ALA", 238, protein=protein)
    mPro = chilife.mutate(protein, D41G, S238A, add_missing_atoms=False)
    D41G_pos = mPro.select_atoms("resnum 41").positions
    S238A_pos = mPro.select_atoms("resnum 238").positions
    np.testing.assert_almost_equal(D41G._coords[0], D41G_pos, decimal=6)
    np.testing.assert_almost_equal(S238A._coords[0], S238A_pos, decimal=6)


def test_mutate5():
    PPII = mda.Universe('test_data/PolyProII.pdb')
    TOC = chilife.SpinLabel('TOC', 8, PPII)

    PPIIm = chilife.mutate(PPII, TOC)
    assert PPIIm.residues[-1].resname == 'NHH'


def test_add_missing_atoms():
    protein = mda.Universe("test_data/1omp.pdb", in_memory=True).select_atoms("protein")
    new_prot = chilife.mutate(protein)
    assert len(new_prot.atoms) != len(protein.atoms)
    assert len(new_prot.atoms) == 2877


@pytest.mark.parametrize(
    "res", set(chilife.dihedral_defs.keys()) -
           {"CYR1", "MTN", "R1M", "R1C"} -
           chilife.USER_dLIBRARIES -
           chilife.USER_LIBRARIES
)
def test_sort_and_internal_coords(res):
    pdbfile = chilife.RL_DIR / f"residue_pdbs/{res.lower()}.pdb"
    lines = chilife.sort_pdb(str(pdbfile))
    anames = [line[13:16] for line in lines if "H" not in line[12:16]]

    with open(pdbfile, "r") as f:
        ans = [
            line[13:16]
            for line in f.readlines()
            if line.startswith("ATOM")
            if "H" not in line[12:16]
        ]

    assert anames == ans


def test_get_min_topol():
    with open('test_data/DHC.pdb', 'r') as f:
        lines = f.readlines()

    all_lines = []
    for line in lines:
        if line.startswith('MODEL'):
            newlines = []
        elif line.startswith(('ATOM', 'HETATOM')):
            newlines.append(line)
        elif line.startswith('ENDMDL'):
            all_lines.append(newlines)

    min_bonds = chilife.get_min_topol(all_lines)
    ans = {(0, 1), (0, 2), (0, 6), (2, 3), (2, 4), (2, 5), (6, 7), (6, 16), (7, 8), (7, 10), (7, 17), (8, 9), (8, 23),
           (10, 11), (10, 18), (10, 19), (11, 12), (11, 13), (12, 15), (12, 20), (13, 14), (13, 21), (14, 15), (14, 22),
           (15, 77), (23, 24), (23, 27), (24, 25), (24, 28), (24, 29), (25, 26), (25, 30), (30, 31), (30, 34), (31, 32),
           (31, 35), (31, 36), (32, 33), (32, 37), (37, 38), (37, 41), (38, 39), (38, 42), (38, 43), (39, 40), (39, 44),
           (44, 45), (44, 54), (45, 46), (45, 48), (45, 55), (46, 47), (46, 61), (48, 49), (48, 56), (48, 57), (49, 50),
           (49, 51), (50, 53), (50, 58), (51, 52), (51, 60), (52, 53), (52, 59), (53, 77), (61, 62), (61, 63), (64, 73),
           (64, 77), (65, 75), (66, 75), (66, 77), (67, 73), (68, 71), (68, 72), (68, 74), (68, 77), (69, 76), (70, 76),
           (71, 73), (71, 78), (71, 79), (72, 76), (72, 82), (72, 83), (74, 75), (74, 80), (74, 81)}

    assert min_bonds == ans


# def test_preferred_dihedrals():
#     dih = [['N', 'CA', 'CB', 'CB2'],
#            ['CA', 'CB', 'CB2', 'CG'],
#            ['ND', 'CE3', 'CZ3', 'C31'],
#            ['CZ1', 'C11', 'C12', 'N12'],
#            ['C11', 'C12', 'N12', 'C13'],
#            ['C12', 'N12', 'C13', 'C14'],
#            ['N12', 'C13', 'C14', 'C15']]
#
#     TEP = mda.Universe('test_data/TEP.pdb')
#     IC = chiLife.get_internal_coords(TEP, resname='TEP', preferred_dihedrals=dih)
#     IC2 = chiLife.get_internal_coords(TEP, resname='TEP')
#
#     IC.get_dihedral(1, dih)
#     with pytest.raises(ValueError):
#         IC2.get_dihedral(1, dih)


from __future__ import annotations
import functools
import numbers
import operator
from functools import partial, update_wrapper
from .protein_utils import sort_pdb
import numpy as np
from numpy.typing import ArrayLike
from scipy.spatial import cKDTree


# TODO:
#   Performance enhancement: Preconstruct Atom objects
#   Behavior: AtomSelections should have orders to be enforced when indexing.
#   Performance enhancement: Find a faster way to retrieve coordinate data from trajectory @property seems to have
#   Feature: Add from_mda class method.

class MolecularSystem:
    """Base class for molecular systems"""

    def __getattr__(self, item):
        if item == 'trajectory':
            return self.protein.trajectory
        else:
            return np.squeeze(self.protein.__getattribute__(item)[self.mask])

    def select_atoms(self, selstr):
        mask = process_statement(selstr, self.logic_keywords, self.protein_keywords)
        if hasattr(self, 'mask'):
            t_mask = np.zeros(self.protein.n_atoms, dtype=bool)
            t_mask[self.mask] = True
            mask *= t_mask

        mask = np.argwhere(mask).T[0]
        return AtomSelection(self.protein, mask)

    @property
    def fname(self):
        return self.protein._fname

    @fname.setter
    def fname(self, val):
        self.protein.fname = val

    @property
    def atoms(self):
        return self.select_atoms("")

    @property
    def positions(self):
        return self.coords

    @positions.setter
    def positions(self, val):
        self.coords = np.asarray(val)

    @property
    def coords(self):
        return self.trajectory.coords[self.trajectory.frame, self.mask]

    @coords.setter
    def coords(self, val):
        self.trajectory.coords[self.trajectory.frame, self.mask] = np.asarray(val)

    @property
    def resindices(self):
        return np.unique(self.protein.resixs[self.mask])

    @property
    def segindices(self):
        return np.unique(self.protein.segixs[self.mask])

    @property
    def n_atoms(self):
        return len(self.mask)

    @property
    def n_residues(self):
        return len(np.unique(self.resnums[self.mask]))

    @property
    def n_chains(self):
        return len(np.unique(self.chains[self.mask]))

    @property
    def residues(self):
        return ResidueSelection(self.protein, self.mask)

    @property
    def segments(self):
        return SegmentSelection(self.protein, self.mask)

    @property
    def logic_keywords(self):
        return self.protein._logic_keywords

    @property
    def protein_keywords(self):
        return self.protein._protein_keywords

    @property
    def types(self):
        return self.protein.atypes[self.mask]

    @property
    def universe(self):
        return self.protein

    def copy(self):
        p2 = self.protein.copy()
        return AtomSelection(p2, self.mask.copy())

    def __iter__(self):
        for idx in self.mask:
            yield self.protein.atoms[idx]


class Protein(MolecularSystem):
    """Protein object"""
    def __init__(
            self,
            atomids: np.ndarray,
            names: np.ndarray,
            altlocs: np.ndarray,
            resnames: np.ndarray,
            resnums: np.ndarray,
            chains: np.ndarray,
            trajectory: np.ndarray,
            occupancies: np.ndarray,
            bs: np.ndarray,
            segs: np.ndarray,
            atypes: np.ndarray,
            charges: np.ndarray,
            name: str = 'Noname_Protein'
    ):
        self.protein = self
        self.atomids = atomids.copy()
        self.names = names.copy()
        self.altlocs = altlocs.copy()
        self.resnames = resnames.copy()
        self.resnums = resnums.copy()
        self.chains = chains.copy()
        self.trajectory = Trajectory(trajectory.copy(), self)
        self.occupancies = occupancies.copy()
        self.bs = bs.copy()
        self.segs = segs.copy()
        self.segids = self.chains
        self.atypes = atypes.copy()
        self.charges = charges.copy()
        self._fname = name

        self.ix = np.arange(len(self.atomids))
        self.mask = np.arange(len(self.atomids))

        resix_borders = np.nonzero(np.r_[1, np.diff(self.resnums)[:-1]])
        resix_borders = np.append(resix_borders, [self.n_atoms])
        resixs = []
        for i in range(len(resix_borders) - 1):
            dif = resix_borders[i + 1] - resix_borders[i]
            resixs.append(np.ones(dif, dtype=int) * i)

        self.resixs = np.concatenate(resixs)
        self.segixs = np.array([ord(x) - 65 for x in self.chains])
        self._Atoms = np.array([Atom(self, i) for i in range(self.n_atoms)])

        self._protein_keywords = {'id': self.atomids,
                                  'name': self.names,
                                  'altloc': self.altlocs,
                                  'resname': self.resnames,
                                  'resnum': self.resnums,
                                  'chain': self.chains,
                                  'occupancies': self.occupancies,
                                  'b': self.bs,
                                  'segid': self.chains,
                                  'type': self.atypes,
                                  'charges': self.charges,

                                  'resi': self.resnums,
                                  'resid': self.resnums,
                                  'i.': self.resnums,
                                  's.': self.chains,
                                  'c.': self.chains,
                                  'r.': self.resnames,
                                  'n.': self.names,
                                  'resn': self.resnames,
                                  'q': self.occupancies,
                                  'elem': self.atypes,
                                  'element': self.atypes,

                                  '_len': self.n_atoms,
                                  }

        self._logic_keywords = {'and': operator.mul,
                                'or': operator.add,
                                'not': unot,
                                '!': unot,
                                '<': operator.lt,
                                '>': operator.gt,
                                '==': operator.eq,
                                '<=': operator.le,
                                '>=': operator.ge,
                                '!=': operator.ne,
                                'byres': partial(byres, protein=self.protein),
                                'within': update_wrapper(partial(within, protein=self.protein), within),
                                'around': update_wrapper(partial(within, protein=self.protein), within)}

    @classmethod
    def from_pdb(cls, file_name):
        """reads a pdb file and returns a Protein object"""

        keys = ["skip", "atomids", "names", "altlocs", "resnames", "chains", "resnums",
                "skip", "coords", "occupancies", "bs", "segs", "atypes", "charges"]

        lines = sort_pdb(file_name)

        if isinstance(lines[0], str):
            lines = [lines]

        PDB_data = [(line[:6].strip(), int(line[6:11]), line[12:16].strip(), line[16:17].strip(),
                     line[17:20].strip(), line[21:22].strip(), int(line[22:26]), line[26:27].strip(),
                     (float(line[30:38]), float(line[38:46]), float(line[46:54])), float(line[54:60]),
                     float(line[60:66]), line[72:73].strip(), line[76:78].strip(), line[78:80].strip())
                    for line in lines[0]]

        pdb_dict = {key: np.array(data) for key, data in zip(keys, zip(*PDB_data)) if key != "skip"}
        trajectory = [pdb_dict.pop('coords')]

        if len(lines) > 1:
            for struct in lines[1:]:
                frame_coords = [(line[30:38], line[38:46], line[46:54]) for line in struct]

                if len(frame_coords) != len(PDB_data):
                    raise ValueError('All models in a multistate PDB must have the same atoms')

                trajectory.append(frame_coords)

        pdb_dict['trajectory'] = np.array(trajectory, dtype=float)
        pdb_dict['name'] = file_name
        return cls(**pdb_dict)

    @classmethod
    def from_arrays(cls,
                    anames: ArrayLike,
                    atypes: ArrayLike,
                    resnames: ArrayLike,
                    resindices: ArrayLike,
                    resnums: ArrayLike,
                    segindices: ArrayLike,
                    segids: ArrayLike = None,
                    trajectory: ArrayLike = None,
                    ) -> Protein:
        anames = np.asarray(anames)
        atypes = np.asarray(atypes)
        resindices = np.asarray(resindices)
        segindices = np.asarray(segindices)

        n_atoms = len(anames)
        atomids = np.arange(n_atoms)
        altlocs = np.array([''] * n_atoms)

        if len(resnums) != n_atoms:
            resnums = np.array([resnums[residx] for residx in resindices])
        else:
            resnums = np.asarray(resnums)

        if len(resnames) != n_atoms:
            resnames = np.array([resnames[residx] for residx in resindices])
        else:
            resnames = np.asarray(resnames)

        if len(segindices) != n_atoms:
            segindices = np.array([segindices[x] for x in resindices])

        if segids is None:
            segids = np.array(["ABCDEFGHIJKLMNOPQRSTUVWXYZ"[i] for i in segindices])
        elif len(segids) != n_atoms:
            segids = np.array([segids[i] for i in segindices])
        else:
            segids = np.asarray(segids)

        chains = segids

        if trajectory is None:
            trajectory = np.empty((1, n_atoms, 3))

        occupancies = np.ones(n_atoms)
        bs = np.ones(n_atoms)
        charges = np.zeros(n_atoms)

        return cls(atomids, anames, altlocs, resnames, resnums, chains,
                   trajectory, occupancies, bs, segids, atypes, charges)

    def copy(self):
        return Protein(
            atomids=self.atomids,
            names=self.names,
            altlocs=self.altlocs,
            resnames=self.resnames,
            resnums=self.resnums,
            chains=self.chains,
            trajectory=self.trajectory.coords,
            occupancies=self.occupancies,
            bs=self.bs,
            segs=self.segs,
            atypes=self.atypes,
            charges=self.charges,
            name=self.names)


class Trajectory:

    def __init__(self, coordinates, protein, timestep=1):
        self.protein = protein
        self.timestep = timestep
        self.coords = coordinates
        self._frame = 0
        self.time = np.arange(0, len(self.coords)) * self.timestep

    def __getitem__(self, item):
        if isinstance(item, (slice, list, np.ndarray)):
            return TrajectoryIterator(self, item)

        elif isinstance(item, numbers.Integral):
            self._frame = item
            return self.time[self._frame]

    def __iter__(self):
        return iter(TrajectoryIterator(self))

    def __len__(self):
        return len(self.time)

    @property
    def frame(self):
        return self._frame

    @frame.setter
    def frame(self, value):
        self._frame = value


class TrajectoryIterator:

    def __init__(self, trajectory, arg=None):
        self.trajectory = trajectory
        if arg is None:
            self.indices = range(len(self.trajectory.time))
        elif isinstance(arg, slice):
            self.indices = range(*arg.indices(len(self.trajectory.time)))
        elif isinstance(arg, (np.ndarray, list, tuple)):
            if np.issubdtype(arg.dtype, np.integer):
                self.indices = arg
            elif np.dtype == bool:
                self.indices = np.argwhere(arg)
            else:
                raise TypeError(f'{arg.dtype} is not a valid indexor for a trajectory')
        else:
            raise TypeError(f'{arg} is not a valid indexor for a trajectory')

    def __iter__(self):
        for frame in self.indices:
            yield self.trajectory[frame]

    def __len__(self):
        return len(self.indices)


def process_statement(statement, logickws, subjectkws):
    sub_statements = parse_paren(statement)
    unary_operators = (logickws['byres'], logickws['not'])
    advanced_operators = (logickws['within'], logickws['around'])

    mask = np.ones(subjectkws['_len'], dtype=bool)
    operation = logickws['and']
    next_operation = None

    for stat in sub_statements:
        cont = False
        stat_split = stat.split()

        if stat_split[0] in logickws:
            if len(stat_split) == 1:
                if logickws[stat_split[0]] in unary_operators:
                    _io = logickws[stat_split[0]]

                    def toperation(a, b, operation, _io):
                        return operation(a, _io(b))

                    operation = functools.partial(toperation, operation=operation, _io=_io)
                else:
                    operation = logickws[stat_split[0]]

                continue

            elif operation in unary_operators:
                _io = logickws[stat_split[0]]
                operation = lambda a, b: operation(_io(a, b))

            elif logickws[stat_split[0]] in unary_operators:
                _io = logickws[stat_split[0]]

                def toperation(a, b, operation, _io):
                    return operation(a, _io(b))

                operation = functools.partial(toperation, operation=operation, _io=_io)

            elif logickws[stat_split[0]] in advanced_operators:
                _io = logickws[stat_split[0]]
                args = [stat_split.pop(i) for i in range(1, 1 + _io.nargs)]
                _io = functools.partial(_io, *args)

                def toperation(a, b, operation, _io):
                    return operation(a, _io(b))

                operation = functools.partial(toperation, operation=operation, _io=_io)

            elif operation != None:
                raise ValueError('Cannot have two logical operators in succession')

            else:
                operation = logickws[stat_split[0]]

            stat_split = stat_split[1:]

        elif stat.startswith('('):
            cont = True
            sub_out = process_statement(stat, logickws, subjectkws)

        if len(stat_split) == 0:
            continue

        if stat_split[0] not in subjectkws and not cont:
            raise ValueError(f"{stat_split[0]} is not a valid keyword. Please start expressions with valid keywords")

        if stat_split[0] in subjectkws:
            finished = False
            sub_out = np.ones(subjectkws['_len'], dtype=bool)
            internal_operation = logickws['and']
            next_internal_operation = None
            while not finished:
                if stat_split[0] in subjectkws:
                    subject = subjectkws[stat_split.pop(0)]
                    values = []

                elif stat_split[0] in logickws:
                    if logickws[stat_split[0]] in unary_operators:
                        _io = logickws[stat_split[0]]

                        def toperation(a, b, operation, _io):
                            return operation(a, _io(b))

                        internal_operation = functools.partial(toperation, operation=internal_operation, _io=_io)
                        stat_split = stat_split[1:]
                        continue
                    else:
                        raise ValueError('Cannot have two logical operators in succession unless the second one is '
                                         '`not`')
                else:
                    raise ValueError(f'{stat_split[0]} is not a valid keyword')

                for i, val in enumerate(stat_split):

                    if logickws.get(val, None) in advanced_operators:
                        _io = logickws[val]
                        args = [stat_split.pop(j) for j in range(i + 1, i + 1 + _io.nargs)]
                        _io = functools.partial(_io, *args)

                        def toperation(a, b, operation, _io):
                            return operation(a, _io(b))

                        next_internal_operation = functools.partial(toperation, operation=operation, _io=_io)
                        stat_split = stat_split[i + 1:]
                        if stat_split == []:
                            next_operation = next_internal_operation
                            next_internal_operation = None
                        break

                    elif val in logickws:
                        next_internal_operation = logickws[val]
                        stat_split = stat_split[i + 1:]
                        if stat_split == []:
                            next_operation = next_internal_operation
                            next_internal_operation = None
                        break
                    elif '-' in val:
                        start, stop = [int(x) for x in val.split('-')]
                        values += list(range(start, stop + 1))
                    elif ':' in val:
                        start, stop = [int(x) for x in val.split(':')]
                        values += list(range(start, stop + 1))
                    else:
                        values.append(val)
                else:
                    finished = True

                values = np.array(values, dtype=subject.dtype)
                tmp = np.isin(subject, values)
                sub_out = internal_operation(sub_out, tmp)

                internal_operation = next_internal_operation
                next_internal_operation = None
                if internal_operation is None:
                    finished = True

        if operation is None:
            print(stat)
            raise ValueError('Need an operation between two selections')

        mask = operation(mask, sub_out)
        operation = next_operation
        next_operation = None

    return mask


def parse_paren(string):
    stack = 0
    start_idx = 0
    end_idx = 0
    results = []

    for i, c in enumerate(string):
        if c == '(':
            if stack == 0:
                start_idx = i + 1
                if i != 0:
                    results.append(string[end_idx:i].strip())
            stack += 1
        elif c == ')':
            # pop stack
            stack -= 1

            if stack == 0:
                results.append(f'({string[start_idx:i].strip()})')
                start_idx = end_idx
                end_idx = i + 1

    if end_idx != len(string) - 1:
        results.append(string[end_idx:].strip())

    results = [res for res in results if res != '']
    if len(results) == 1 and results[0].startswith('('):
        results = parse_paren(results[0][1:-1])

    return results


class AtomSelection(MolecularSystem):

    def __new__(cls, protein, mask):
        if isinstance(mask, int):
            return Atom(protein, mask)
        else:
            return object.__new__(cls)

    def __init__(self, protein, mask):
        self.protein = protein
        self.mask = mask

    def __getitem__(self, item):

        if np.issubdtype(type(item), np.integer):
            relidx = self.protein.ix[self.mask][item]
            return Atom(self.protein, relidx)

        if isinstance(item, slice):
            return AtomSelection(self.protein, self.mask[item])

        elif isinstance(item, (np.ndarray, list, tuple)):
            item = np.asarray(item)
            if len(item) == 1 or item.sum() == 1:
                return Atom(self.protein, self.mask[item])
            elif item.dtype == bool:
                item = np.argwhere(item).T[0]
                return AtomSelection(self.protein, item)
            else:
                return AtomSelection(self.protein, self.mask[item])

        elif hasattr(item, '__iter__'):
            if all([np.issubdtype(type(x), int) for x in item]):
                return AtomSelection(self.protein, self.mask[item])

        raise TypeError('Only integer, slice type, and boolean mask arguments are supported at this time')

    def __len__(self):
        return len(self.mask)


class ResidueSelection(MolecularSystem):

    def __init__(self, protein, mask):

        resixs = np.unique(protein.resixs[mask])
        self.mask = np.isin(protein.resixs, resixs)
        self.protein = protein

        first_ix = np.nonzero(np.r_[1, np.diff(protein.resixs)[:-1]])[0]
        self.first_ix = np.array([ix for ix in first_ix if
                                  np.isin(protein.resixs[ix], protein.resixs[self.mask])],
                                 dtype=int)

        self.resnames = protein.resnames[self.first_ix].flatten()
        self.resnums = protein.resnums[self.first_ix].flatten()
        self.segids = protein.segids[self.first_ix].flatten()
        self.chains = protein.chains[self.first_ix].flatten()

    def __getitem__(self, item):
        resixs = np.unique(self.protein.resixs[self.mask])
        new_resixs = resixs[item]
        new_mask = np.argwhere(np.isin(self.protein.resixs, new_resixs))

        if np.issubdtype(type(item), int):
            return Residue(self.protein, new_mask)
        elif isinstance(item, slice):
            return ResidueSelection(self.protein, new_mask)
        elif hasattr(item, '__iter__'):
            if all([np.issubdtype(type(x), int) for x in item]):
                return ResidueSelection(self.protein, new_mask)

        raise TypeError('Only integer and slice type arguments are supported at this time')

    def __len__(self):
        return len(self.first_ix)

    def __iter__(self):
        for resnum in self.resnums:
            mask = self.protein.resnums == resnum
            yield Residue(self.protein, mask)


class SegmentSelection(MolecularSystem):

    def __init__(self, protein, mask):
        seg_ixs = np.unique(protein.segixs[mask])
        self.mask = np.argwhere(np.isin(protein.segixs, seg_ixs)).T[0]
        self.protein = protein

        first_ix = np.nonzero(np.r_[1, np.diff(protein.segixs)[:-1]])
        self.first_ix = np.array([ix for ix in first_ix if protein.segixs[ix] in protein.segixs[self.mask]])

        self.segids = protein.segids[self.first_ix].flatten()
        self.chains = protein.chains[self.first_ix].flatten()

    def __getitem__(self, item):
        segixs = np.unique(self.protein.segixs[self.mask])
        new_segixs = segixs[item]
        new_mask = np.argwhere(np.isin(self.protein.segixs, new_segixs))

        if np.issubdtype(type(item), int):
            return Segment(self.protein, new_mask)
        elif isinstance(item, slice):
            return SegmentSelection(self.protein, new_mask)
        elif hasattr(item, '__iter__'):
            if all([np.issubdtype(type(x), int) for x in item]):
                return SegmentSelection(self.protein, new_mask)

        raise TypeError('Only integer and slice type arguments are supported at this time')

    def __len__(self):
        return len(self.first_ix)


class Atom(MolecularSystem):
    def __init__(self, protein, mask):
        self.index = mask
        self.mask = mask

        self.protein = protein
        self.name = protein.names[self.index]
        self.altLoc = protein.altlocs[self.index]
        self.atype = protein.types[self.index]
        self.type = self.atype

        self.resn = protein.resnames[self.index]
        self.resname = self.resn
        self.resi = protein.resnums[self.index]
        self.resid = self.resi
        self.resnum = self.resi
        self.chain = protein.segids[self.index]
        self.segid = protein.chains[self.index]

    @property
    def position(self):
        return self.protein.coords[self.index]


class Residue(MolecularSystem):
    def __init__(self, protein, mask):
        resix = np.unique(protein.resixs[mask])[0]
        self.mask = np.argwhere(np.isin(protein.resixs, resix)).T[0]
        self.protein = protein

        self.resname = protein.resnames[self.mask][0]
        self.resnum = protein.resnums[self.mask][0]
        self.resid = self.resnum
        self.segid = protein.segids[self.mask][0]
        self.segindex = protein.segixs[self.mask][0]
        self.chain = protein.chains[self.mask][0]

    def __len__(self):
        return len(self.mask)

    def phi_selection(self):
        prev = self.atoms.resnums - 1
        prev = np.unique(prev[prev > 0])
        resnums = np.unique(self.atoms.resnums)
        sel = self.protein.select_atoms(f"(resnum {' '.join(str(r) for r in resnums)} and name N CA C) or "
                                        f"(resnum {' '.join(str(r) for r in prev)} and name C))")
        return sel if len(sel) == 4 else None

    def psi_selection(self):
        nex = self.atoms.resnums + 1
        nex = np.unique(nex[nex <= self.protein.resnums.max()])
        resnums = np.unique(self.resnums)
        sel = self.protein.select_atoms(f"(resnum {' '.join(str(r) for r in resnums)} and name N CA C) or "
                                        f"(resnum {' '.join(str(r) for r in nex)} and name N))")
        return sel if len(sel) == 4 else None


class Segment(MolecularSystem):

    def __init__(self, protein, mask):
        segix = np.unique(protein.segixs[mask])[0]
        self.mask = np.argwhere(np.isin(protein.segixs, segix)).T[0]
        self.protein = protein

        self.segid = protein.segids[segix]
        self.chain = protein.chains[segix]


def byres(mask, protein):
    residues = np.unique(protein.resnums[mask])
    mask = np.isin(protein.resnums, residues)
    return mask


def unot(mask):
    return ~mask


def within(distance, mask, protein):
    distance = float(distance)
    tree1 = cKDTree(protein.coords)
    tree2 = cKDTree(protein.protein.coords[mask])
    results = np.concatenate(tree2.query_ball_tree(tree1, distance))
    results = np.unique(results)
    out_mask = np.zeros_like(mask, dtype=bool)
    out_mask[results] = True
    out_mask = out_mask * ~mask
    return out_mask


within.nargs = 1
